from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Generator, NamedTuple, TypeVar

from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition

from .density_matrix import DensityMatrix
from ..utils import Logger, add_fake_kpts, two_communicators
from ..utils.memory import HasMemoryEstimate, MemoryEstimate
from ..typing import Communicator


class WorkMetadata(NamedTuple):
    """ Metadata to the density matrices """
    density_matrices: BaseDensityMatrices

    @property
    def global_indices(self) -> tuple[int, ...]:
        """ Unique index for this work. """
        raise NotImplementedError

    @property
    @abstractmethod
    def desc(self) -> str:
        raise NotImplementedError

    def __str__(self) -> str:
        return f'{self.__class__.__name__}{self.global_indices}'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}{self.global_indices}'


WorkMetadataT = TypeVar('WorkMetadataT', bound=WorkMetadata)


class BaseDensityMatrices(HasMemoryEstimate, ABC, Generic[WorkMetadataT]):

    _log: Logger
    _ksd: KohnShamDecomposition

    """
    Collection of density matrices in the Kohn-Sham basis for different times
    or frequencies, possibly after convolution with various pulses.

    Plain density matrices and/or derivatives thereof may be represented.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    real
        Calculate the real part of density matrices.
    imag
        Calculate the imaginary part of density matrices.
    calc_size
        Size of the calculation communicator.
    """

    def __init__(self,
                 ksd: KohnShamDecomposition | str,
                 real: bool = True,
                 imag: bool = True,
                 calc_size: int = 1,
                 log: Logger | None = None):
        assert real or imag
        self._reim_r: list[str] = []
        if real:
            self._reim_r.append('Re')
        if imag:
            self._reim_r.append('Im')

        if log is None:
            self._log = Logger()
        else:
            self._log = log

        self._loop_comm, self._calc_comm = two_communicators(-1, calc_size)
        if isinstance(ksd, KohnShamDecomposition):
            self._ksd = ksd
        else:
            self._ksd = KohnShamDecomposition(filename=ksd)
        add_fake_kpts(self._ksd)

        # Do a quick sanity check at runtime
        self._runtime_verify_work_loop()

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError

    def get_memory_estimate(self) -> MemoryEstimate:
        memory_estimate = MemoryEstimate(comment='Unknown')

        return memory_estimate

    def parallel_prepare(self):
        """ Read everything necessary synchronously on all ranks. """

    @abstractmethod
    def __iter__(self) -> Generator[tuple[WorkMetadataT, DensityMatrix], None, None]:
        """ Obtain density matrices for various times, pulses or frequencies.

        Yields
        ------
        Tuple (work, dm) on the root rank of the calculation communicator:

            work
                An object representing the metadata (time, frequency or pulse) for the work done.
            dm
                Density matrix for this time, frequency or pulse.
        """
        raise NotImplementedError

    def iread_gather_on_root(self) -> Generator[tuple[WorkMetadataT, DensityMatrix], None, None]:
        """ Obtain density matrices for various times, pulses or frequencies and gather to the root rank.

        Yields
        ------
        Tuple (work, dm) on the root rank of the loop and calculation communicators:

            work
                An object representing the metadata (time, frequency or pulse) for the work done.
            dm
                Density matrix for this time, frequency or pulse.
        """
        work: WorkMetadataT | None
        gen = iter(self)

        # Loop over the work to be done, and the ranks that are supposed to do it
        self.parallel_prepare()
        for rank, work in self.global_work_loop_with_idle():
            if work is None:
                # Rank rank will not do any work at this point
                continue

            if rank == self.loop_comm.rank:
                mywork, mydm = next(gen)
                if self.calc_comm.rank == 0:
                    self.log(f'Read {mywork.desc} in {self.log.elapsed("read"):.1f}s',
                             who='Response', if_elapsed=5)
                assert work.global_indices == mywork.global_indices, f'{work.desc} != {mywork.desc}'

            dm = DensityMatrix.broadcast(
                mydm if self.loop_comm.rank == rank else None,
                ksd=self.ksd,
                root=rank, dm_comm=self.calc_comm, comm=self.loop_comm)

            yield work, dm

        _exhausted = object()
        rem = next(gen, _exhausted)
        assert rem is _exhausted, rem

    @property
    def ksd(self) -> KohnShamDecomposition:
        """ Kohn-Sham decomposition object. """
        return self._ksd

    @property
    def log(self) -> Logger:
        """ Logger. """
        return self._log

    def log_parallel(self, *args, **kwargs) -> Logger:
        """ Log message with communicator information. """
        return self._log(*args, **kwargs, comm=self.loop_comm, who='Response')

    @property
    def reim(self) -> list[str]:
        """ List of strings ``'Re'`` and ``'Im'``, depending on whether real, and/or imaginary parts are computed. """
        return self._reim_r

    @abstractmethod
    def work_loop(self,
                  rank: int) -> Generator[WorkMetadataT | None, None, None]:
        """ The work to be done by a certain rank of the loop communicator.

        Parameters
        ----------
        rank
            Rank of the loop communicator.

        Yields
        ------
        Objects representing the time, frequency or pulse to be computed by rank ``rank``.
        None is yielded when `rank` does not do any work while other ranks are doing work.
        """
        raise NotImplementedError

    def _runtime_verify_work_loop(self):
        """ Verify that the description of work to be done is consistent across ranks. """
        local_work_r = [list(self.work_loop(rank)) for rank in range(self.loop_comm.size)]
        work_lengths = [len(local_work) for local_work in local_work_r]
        assert all([work_lengths[0] == work_length for work_length in work_lengths]), \
            f'The work loop has different length across the different ranks. {work_lengths}'
        concat_work_list = [work.global_indices for local_work in local_work_r for work in local_work
                            if work is not None]
        assert len(concat_work_list) == len(set(concat_work_list)), \
            f'Different ranks do duplicate work {concat_work_list}'

    @property
    def local_work_plan(self) -> tuple[WorkMetadataT, ...]:
        """ The work to be done by a this rank of the loop communicator.

        Yields
        ------
        Objects representing the time, frequency or pulse to be computed by this rank.
        """
        local_work_plan = tuple(work for work in self.work_loop(self.loop_comm.rank)
                                if work is not None)
        return local_work_plan

    @property
    def local_work_plan_with_idle(self) -> tuple[WorkMetadataT | None, ...]:
        """ The work to be done by a this rank of the loop communicator.

        This function includes None values for when this rank does not do any work
        in order to synchronize the execution.

        Yields
        ------
        Objects representing the time, frequency or pulse to be computed by this rank.
        None is yielded when this rank does not do any work while other ranks are doing work.
        """
        local_work_plan = tuple(self.work_loop(self.loop_comm.rank))

        return local_work_plan

    def global_work_loop_with_idle(self) -> Generator[tuple[int, WorkMetadataT | None], None, None]:
        """ The work to be done by a all ranks of the loop communicator.

        This function includes None values for when ranks do not do any work
        in order to synchronize the execution.

        Yields
        ------
        Lists of length equal to the loop communicator size. Each element in the list represents
        the work to be done. See `local_work_plan_with_idle`.
        """
        work_loop_r = [self.work_loop(rank) for rank in range(self.loop_comm.size)]
        while True:
            for rank in range(self.loop_comm.size):
                try:
                    work = next(work_loop_r[rank])
                    yield rank, work
                except StopIteration:
                    if rank == 0:
                        # No more work to do
                        return
                    else:
                        raise RuntimeError(f'Ranks have different amount of work. Exited on {rank}')

    def global_work_loop(self) -> Generator[tuple[int, WorkMetadataT | None], None, None]:
        """ The work to be done by a all ranks of the loop communicator.

        Yields
        ------
        Lists of length equal to the loop communicator size. Each element in the list represents
        the work to be done. See :func:`local_work_plan`.
        """
        for rank, work in self.global_work_loop_with_idle():
            if work is None:
                continue
            yield rank, work

    @property
    def localn(self) -> int:
        """ Total number of density matrices this rank will work with. """
        return len(self.local_work_plan)

    @property
    def globaln(self) -> int:
        """ Total number of density matrices to work with across all ranks. """
        local_work_r = [list(self.work_loop(rank)) for rank in range(self.loop_comm.size)]
        concat_work_list = [work for local_work in local_work_r for work in local_work
                            if work is not None]
        return len(concat_work_list)

    @property
    def calc_comm(self) -> Communicator:
        """ Calculation communicator.

        Each rank of this communicator calculates the observables corresponding to
        a part (in electron-hole space) of the density matrices. """
        return self._calc_comm

    @calc_comm.setter
    def calc_comm(self, value: Communicator):
        from gpaw.mpi import world
        if value is None:
            self.calc_comm = world
            return

        assert hasattr(value, 'rank')
        assert hasattr(value, 'size')
        self._calc_comm = value

    @property
    def loop_comm(self) -> Communicator:
        """ Loop communicator.

        Each rank of this communicator calculates the density matrices corresponding to
        different times, frequencies or after convolution with a different pulse. """
        return self._loop_comm

    @loop_comm.setter
    def loop_comm(self, value: Communicator):
        from gpaw.mpi import world
        if value is None:
            self.loop_comm = world
            return

        assert hasattr(value, 'rank')
        assert hasattr(value, 'size')
        self._loop_comm = value
        raise NotImplementedError

    @abstractmethod
    def write_to_disk(self,
                      fmt: str):
        """ Calculate the density matrices amd save to disk.

        Parameters
        ----------
        fmt
            Formatting string.
        """
        raise NotImplementedError



from __future__ import annotations

from typing import Generator, Generic, Sequence
import numpy as np
from numpy.typing import NDArray
from numpy._typing import _DTypeLike as DTypeLike  # parametrizable wrt generic

from ..utils import DTypeT, Logger, env
from ..typing import ArrayIndex


class DensityMatrixBuffer(Generic[DTypeT]):

    """ Buffer for the density matrix.

    Objects of this class can hold buffers for real and imaginary parts
    and various derivatives at the same time.

    Each buffer has two dimensions corresponding to (part of) the
    density matrix, and optionally additional dimensions for e.g. time.

    Parameters
    ----------
    nnshape
        Shape of the dimension corresponding to the density matrix. Must
        have dimension 2.
    xshape
        Shape of the additional dimension corresponding to, e.g., time.
    dtype
        Data type of the density matrices.
    re_buffers
        Buffers for the different derivatives of the real part of the
        density matrix as dictionaries, where the keys is the derivative
        order (0, 1, 2, etc.) and the value is a numpy array of shape
        ``(nnshape, xshape)``.
    im_buffers
        Same as :attr:`re_buffers` but for imaginary part.
    """

    def __init__(self,
                 nnshape: tuple[int, int],
                 xshape: tuple[int, ...],
                 dtype: DTypeLike[DTypeT],
                 re_buffers: dict[int, NDArray[DTypeT]] = dict(),
                 im_buffers: dict[int, NDArray[DTypeT]] = dict()):
        assert len(nnshape) == 2
        assert all(isinstance(dim, (int, np.integer)) and dim >= 0 for dim in nnshape)
        assert all(isinstance(dim, (int, np.integer)) and dim >= 0 for dim in xshape)
        assert isinstance(np.dtype(dtype), np.dtype)
        self._nnshape = nnshape
        self._xshape = xshape
        self._dtype = np.dtype(dtype)
        self._re_buffers: dict[int, NDArray[DTypeT]] = dict()
        self._im_buffers: dict[int, NDArray[DTypeT]] = dict()

        for derivative, buffer_nnx in re_buffers.items():
            self.store(True, derivative, buffer_nnx)

        for derivative, buffer_nnx in im_buffers.items():
            self.store(False, derivative, buffer_nnx)

    @property
    def real(self) -> NDArray[DTypeT]:
        """ Buffer of shape nnshape + xshape corresponding to real data. """
        return self._get_real(0)

    @property
    def real1(self) -> NDArray[DTypeT]:
        """ Buffer of shape nnshape + xshape corresponding to real part of first derivative. """
        return self._get_real(1)

    @property
    def real2(self) -> NDArray[DTypeT]:
        """ Buffer of shape nnshape + xshape corresponding to real part of second derivative. """
        return self._get_real(2)

    @property
    def imag(self) -> NDArray[DTypeT]:
        """ Buffer of shape nnshape + xshape corresponding to imaginary data. """
        return self._get_imag(0)

    @property
    def imag1(self) -> NDArray[DTypeT]:
        """ Buffer of shape nnshape + xshape corresponding to imaginary part of first derivative. """
        return self._get_imag(1)

    @property
    def imag2(self) -> NDArray[DTypeT]:
        """ Buffer of shape nnshape + xshape corresponding to imaginary part of second derivative. """
        return self._get_imag(2)

    def _get_real(self,
                  derivative: int) -> NDArray[DTypeT]:
        """ Fetch density matrix buffer for real data.

        Parameters
        ----------
        derivative
            Derivative order.
        """
        return self._re_buffers[derivative]

    def _get_imag(self,
                  derivative: int) -> NDArray[DTypeT]:
        """ Fetch density matrix buffer for imaginary data.

        Parameters
        ----------
        derivative
            Derivative order.
        """
        return self._im_buffers[derivative]

    def _get_data(self,
                  real: bool,
                  derivative: int) -> NDArray[DTypeT]:
        """ Fetch density matrix buffer.

        Parameters
        ----------
        real
            ``True`` if real, ``False`` if imaginary.
        derivative
            Derivative order.
        """
        return self._get_real(derivative) if real else self._get_imag(derivative)

    def copy(self) -> DensityMatrixBuffer:
        """ Return a deep copy of this object (buffers are copied). """
        re_buffers = {derivative: np.array(buffer_nnx)
                      for derivative, buffer_nnx in self._re_buffers.items()}
        im_buffers = {derivative: np.array(buffer_nnx)
                      for derivative, buffer_nnx in self._im_buffers.items()}

        dm_buffer = DensityMatrixBuffer(self.nnshape, self.xshape,
                                        dtype=self.dtype,
                                        re_buffers=re_buffers,
                                        im_buffers=im_buffers)
        return dm_buffer

    def new(self) -> DensityMatrixBuffer:
        """ Return a new buffer with the same shape. """
        dm_buffer = DensityMatrixBuffer(self.nnshape, self.xshape,
                                        dtype=self.dtype)
        return dm_buffer

    def __getitem__(self,
                    value) -> DensityMatrixBuffer:
        """ Index the buffers and return a new DensityMatrixBuffer
        with buffers that are views of the buffers of this DensityMatrixBuffer.
        """
        if len(self._im_buffers) == 0 and len(self._re_buffers) == 0:
            # This case needs some special handing to get the dimension of
            # the output
            raise NotImplementedError

        # Wrap in a tuple
        if not isinstance(value, tuple):
            value = (value, )
        s = (slice(None), slice(None)) + value
        re_buffers = {derivative: buffer_nnx[s]
                      for derivative, buffer_nnx in self._re_buffers.items()}
        im_buffers = {derivative: buffer_nnx[s]
                      for derivative, buffer_nnx in self._im_buffers.items()}

        # Ugly hack. Get any of the buffers
        xshape = (list(re_buffers.values()) + list(im_buffers.values()))[0].shape[2:]
        return DensityMatrixBuffer(self.nnshape, xshape, dtype=self.dtype,
                                   re_buffers=re_buffers, im_buffers=im_buffers)

    @property
    def narrays(self) -> int:
        """ Number of arrays stored in this object. """
        return len(self.derivatives_imag) + len(self.derivatives_real)

    @property
    def nnshape(self) -> tuple[int, int]:
        """ Shape of the part of the density matrix. """
        return self._nnshape

    @property
    def xshape(self) -> tuple[int, ...]:
        """ Shape of the additional dimension of the buffers. """
        return self._xshape

    @property
    def shape(self) -> tuple[int, ...]:
        """ Shape of the buffers. """
        return self.nnshape + self.xshape

    @property
    def dtype(self) -> np.dtype[DTypeT]:
        """ Dtype of the buffers. """
        return self._dtype

    def store(self,
              real: bool,
              derivative: int,
              buffer_nnx: NDArray[DTypeT]):
        """ Store buffer for part of density matrix.

        Parameters
        ----------
        real
            ``True`` if real, ``False`` if imaginary.
        derivative
            Derivative order.
        buffer_nnx
            Buffer of shape ``(nnshape, xshape)``.
        """
        assert isinstance(derivative, int) and derivative >= 0, derivative
        assert isinstance(buffer_nnx, np.ndarray)
        assert buffer_nnx.shape == self.shape, f'{buffer_nnx.shape} != {self.shape}'
        assert buffer_nnx.dtype == self.dtype
        if real:
            self._re_buffers[derivative] = buffer_nnx
        else:
            self._im_buffers[derivative] = buffer_nnx

    def zero_buffers(self,
                     real: bool,
                     imag: bool,
                     derivative_order_s: list[int]):
        """ Initialize many buffers at once to and write zeros.

        Parameters
        ----------
        real
            Initialize buffers for real parts.
        imag
            Initialize buffers for imaginary parts.
        derivative_order_s
            Initialize buffers for these derivative orders.
        """
        for derivative in derivative_order_s:
            if real:
                self.zeros(True, derivative)
            if imag:
                self.zeros(False, derivative)

    def zeros(self,
              real: bool,
              derivative: int):
        """ Initialize buffer with zeros for part of density matrix.

        Parameters
        ----------
        real
            ``True`` if real, ``False`` if imaginary.
        derivative
            Derivative order.
        """
        self.store(real, derivative, np.zeros(self.shape, dtype=self.dtype))

    def broadcast_x(self,
                    data_nnx: NDArray[DTypeT]) -> NDArray[DTypeT]:
        """ Broadcast the x dimensions of data_nnx. """
        nnshape = data_nnx.shape[:2]
        data_xnn = np.moveaxis(np.moveaxis(data_nnx, 0, -1), 0, -1)
        data_xnn = np.broadcast_to(data_xnn, self.xshape + nnshape)
        data_nnx = np.moveaxis(np.moveaxis(data_xnn, -1, 0), -1, 0)
        return data_nnx

    @property
    def nnellipsis(self) -> tuple[slice, slice]:
        return (slice(None), slice(None))

    @property
    def xellipsis(self) -> tuple[slice, ...]:
        return tuple(len(self.xshape) * [slice(None)])

    def safe_fill(self,
                  real: bool,
                  derivative: int,
                  data_nnx: NDArray[DTypeT]):
        """ Write data_nnx to the corrsponding buffer, if the dimensions of data_nnx
        are equal to or smaller than the buffer.

        If the dimensions of data_nnx are smaller than or equal to the dimensions
        of the buffer, write to the first elements of the buffer.
        Otherwise raise and error.

        Parameters
        ----------
        real
            ``True`` if real, ``False`` if imaginary.
        derivative
            Derivative order.
        buffer_nnx
            Data of shape ``(nnshape, xshape)``.
        """
        assert len(data_nnx.shape) <= len(self.shape), f'{data_nnx.shape} > {self.shape}'
        assert all([dima >= dimb for dima, dimb in zip(self.nnshape, data_nnx.shape[:2])]), \
            f'{self.nnshape} < {data_nnx.shape[:2]}'
        data_nnx = self.broadcast_x(data_nnx)  # Broadcast the last dimensions
        assert data_nnx.shape[2:] == self.xshape, f'{data_nnx.shape[2:]} != {self.xshape}'
        s = tuple([slice(dim) for dim in data_nnx.shape[:2]]) + self.xellipsis
        buffer_nnx = self._get_data(real, derivative)
        buffer_nnx[s] = data_nnx

    def safe_add(self,
                 real: bool,
                 derivative: int,
                 data_nnx: NDArray[DTypeT]):
        """ Add data_nnx to the corrsponding buffer, if the dimensions of data_nnx
        are equal to or smaller than the buffer

        If the dimensions of data_nnx are smaller than or equal to the dimensions
        of the buffer, add to the first elements of the buffer.
        Otherwise raise and error.

        Parameters
        ----------
        real
            ``True`` if real, ``False`` if imaginary.
        derivative
            Derivative order.
        buffer_nnx
            Data of shape ``(nnshape, xshape)``.
        """
        assert len(data_nnx.shape) <= len(self.shape), f'{data_nnx.shape} > {self.shape}'
        assert all([dima >= dimb for dima, dimb in zip(self.nnshape, data_nnx.shape[:2])]), \
            f'{self.nnshape} < {data_nnx.shape[:2]}'
        data_nnx = self.broadcast_x(data_nnx)  # Broadcast the last dimensions
        assert data_nnx.shape[2:] == self.xshape, f'{data_nnx.shape[2:]} != {self.xshape}'
        s = tuple([slice(dim) for dim in data_nnx.shape[:2]]) + self.xellipsis
        buffer_nnx = self._get_data(real, derivative)
        # Regarding ignore:
        # https://stackoverflow.com/questions/74633074/how-to-type-hint-a-generic-numpy-array/74634650#74634650
        buffer_nnx[s] += data_nnx  # type: ignore

    @property
    def derivatives_real(self) -> list[int]:
        """ Stored derivative order of real density matrices in sorted order """
        return list(sorted(self._re_buffers.keys()))

    @property
    def derivatives_imag(self) -> list[int]:
        """ Stored derivative order of real density matrices in sorted order """
        return list(sorted(self._im_buffers.keys()))

    def _iter_buffers(self) -> Generator[NDArray[DTypeT], None, None]:
        """ Loop over all real imaginary buffers in a sorted order """
        for derivative in self.derivatives_real:
            yield self._re_buffers[derivative]
        for derivative in self.derivatives_imag:
            yield self._im_buffers[derivative]

    def _iter_reim_derivatives(self) -> Generator[tuple[bool, int], None, None]:
        """ Loop over tuples (real, derivative) in sorted order.

        The parameter real is ``True`` for real buffers and the parameter derivative denotes the
        derivative order of the buffer.
        """
        for derivative in self.derivatives_real:
            yield (True, derivative)
        for derivative in self.derivatives_imag:
            yield (False, derivative)

    def ensure_contiguous_buffers(self):
        """ Make the buffers contiguous if they are not already. """
        for derivative in self.derivatives_real:
            self._re_buffers[derivative] = np.ascontiguousarray(self._re_buffers[derivative])
        for derivative in self.derivatives_imag:
            self._im_buffers[derivative] = np.ascontiguousarray(self._im_buffers[derivative])

    def send_arrays(self,
                    comm,
                    rank: int,
                    log: Logger | None = None):
        """ Send data to another MPI rank.

        Parameters
        ----------
        comm
            Communicator.
        rank
            Send to this rank.
        log
            Optional logger.
        """
        if comm.rank == rank:
            # Sending to send
            return

        if log is not None:
            log.start('send_to_root')

        requests = []
        for mpitag, buffer_nnx in enumerate(self._iter_buffers(), start=987):
            buffer_nnx = np.ascontiguousarray(buffer_nnx)
            requests.append(comm.send(buffer_nnx, 0, tag=mpitag, block=False))
        comm.waitall(requests)

        if log is not None:
            log(f'Sending to root {log.elapsed("send_to_root"):.1f}s', who='Response', flush=True)

    def recv_arrays(self,
                    comm,
                    rank: int,
                    log: Logger | None = None):
        """ Receive data to another MPI rank.

        Parameters
        ----------
        comm
            Communicator.
        rank
            Send to this rank.
        log
            Optional logger.
        """
        if comm.rank == rank:
            # Receiving from self
            return

        if log is not None:
            log.start('root_recv')

        requests = []
        for mpitag, buffer_nnx in enumerate(self._iter_buffers(), start=987):
            requests.append(comm.receive(buffer_nnx, rank, tag=mpitag, block=False))
        comm.waitall(requests)

        if log is not None:
            log(f'Root received {log.elapsed("root_recv"):.1f}s from {rank}', who='Response', flush=True)

    def redistribute(self,
                     target: DensityMatrixBuffer,
                     comm,
                     source_indices_r: Sequence[tuple[ArrayIndex, ...] | ArrayIndex | None],
                     target_indices_r: Sequence[tuple[ArrayIndex, ...] | ArrayIndex | None],
                     log: Logger | None = None,
                     ):
        """ Redistribute this DensityMatrixBuffer to another according to user specified options.

        The nn dimensions of the self and target buffers should be the same,
        but the x dimensions can be different. However, self need to have the same shape on all ranks
        and target needs to have the same shape on all ranks.

        Parameters
        ----------
        target
            Target :class:`DensityMatrixBuffer`.
        comm
            MPI communicator.
        source_indices_r
            List of x-indices. The length of the list must equal to the communicator size.
            The x-index that is element r of the list corresponds
            to the data from the source that will be sent to rank r.
            If the x-index is None, then the rank corresponding to that element will not be
            receiving data.  This argument must be the same on all ranks
        recv_indices_r
            List of x-indices. The length of the list must equal to the communicator size.
            The x-index that is element r of the list corresponds
            to the data in the target that will be received from rank r.
            If the x-index is None, then the rank corresponding to that element will not be
            sending data.  This argument must be the same on all ranks
        log
            Optional logger.
        """
        # Size of each density matrix (the nn-dimensions)
        nnsize = int(np.prod(self.nnshape))
        # Convert maxsize to maximum number of elements
        maxsize = env.get_float('REDISTRIBUTE_MAXSIZE')
        maxtotalelems = int(np.ceil(maxsize / self.dtype.itemsize))

        assert len(source_indices_r) == comm.size, len(source_indices_r)
        assert len(target_indices_r) == comm.size, len(target_indices_r)

        # Extract source and target indices that are not None and make sure they are tuples
        source_indices_by_rank = {rank: x_indices if isinstance(x_indices, tuple) else (x_indices, )
                                  for rank, x_indices in enumerate(source_indices_r)
                                  if x_indices is not None}
        target_indices_by_rank = {rank: x_indices if isinstance(x_indices, tuple) else (x_indices, )
                                  for rank, x_indices in enumerate(target_indices_r)
                                  if x_indices is not None}
        assert len(source_indices_by_rank) > 0
        assert len(target_indices_by_rank) > 0

        # Make sure that same derivatives and real/imaginary parts are stored and that dtypes are same
        assert tuple(self.derivatives_real) == tuple(target.derivatives_real)
        assert tuple(self.derivatives_imag) == tuple(target.derivatives_imag)
        assert self.dtype == target.dtype, f'{self.dtype} != {target.dtype}'

        # Get the xshapes of all sources
        source_xshape_by_rank: dict[int, tuple[int, ...]] = dict()
        if comm.rank in target_indices_by_rank.keys():
            for buf_nnx in self._iter_buffers():
                source_xshape_by_rank = {rank: buf_nnx[self.nnellipsis + x_indices].shape[2:]
                                         for rank, x_indices in source_indices_by_rank.items()}
                break

        # Get the xshapes of the targets by an alltoall operation with the sources
        # -2 means nothing, -1 in first field means empty tuple
        xdims = max(len(self.xshape), len(target.xshape))
        pad_target_xshape_r = -2 * np.ones((comm.size, xdims), dtype=int)
        pad_source_xshape_r = -2 * np.ones((comm.size, xdims), dtype=int)
        for rank, xshape in source_xshape_by_rank.items():
            pad_source_xshape_r[rank, :len(xshape)] = xshape
            if xshape == ():
                pad_source_xshape_r[rank, 0] = -1
        comm.alltoallv(pad_source_xshape_r,
                       xdims * np.ones(comm.size, dtype=int),
                       xdims * np.arange(comm.size, dtype=int),
                       pad_target_xshape_r,
                       xdims * np.ones(comm.size, dtype=int),
                       xdims * np.arange(comm.size, dtype=int))
        target_xshape_by_rank = {rank: tuple(xshape[xshape > -1])
                                 for rank, xshape in enumerate(pad_target_xshape_r) if xshape[0] > -2}
        # Check that target sizes supplied by the user are shorter than
        # or equal to the sizes from the alltoall operation
        if comm.rank in source_indices_by_rank.keys():
            for buf_nnx in target._iter_buffers():
                for rank, x_indices in target_indices_by_rank.items():
                    target_xshape = target_xshape_by_rank[rank]
                    xshape = buf_nnx[self.nnellipsis + x_indices].shape[2:]
                    assert all(np.less_equal(target_xshape, xshape))
                break

        # Get the total number of density matrices that this rank sends/receives
        source_xsize_by_rank = {rank: np.prod(xshape, dtype=int)
                                for rank, xshape in source_xshape_by_rank.items()}
        target_xsize_by_rank = {rank: np.prod(xshape, dtype=int)
                                for rank, xshape in target_xshape_by_rank.items()}
        my_sourcexsize = sum(source_xsize_by_rank.values())
        my_targetxsize = sum(target_xsize_by_rank.values())
        # Get the total number of array elements to be sent across all ranks
        sizes = np.array([my_sourcexsize, my_targetxsize], dtype=int)
        comm.sum(sizes, root=-1)
        total_sourcexsize, total_targetxsize = sizes
        totalsize = max(total_sourcexsize, total_targetxsize) * nnsize

        # Split the data across the nn-dimensions since they are always the same; how many times?
        factortoolarge = totalsize / maxtotalelems
        nnstride = int(np.ceil(nnsize / factortoolarge))
        nnstride = min(nnsize, nnstride)
        nsplits = int(np.ceil(nnsize / nnstride))
        if log is not None and comm.rank == 0:
            total_MiB = totalsize * self.dtype.itemsize / (1024 ** 2)
            buftotal_MiB = totalsize / nsplits * self.dtype.itemsize / (1024 ** 2)

            log(f'Redistribute: {len(target_indices_by_rank)} sending '
                f'and {len(source_indices_by_rank)} receiving. '
                f'Total size on all ranks ({total_MiB:.1f} MiB) '
                f'splitting in {nsplits} parts ({buftotal_MiB:.1f} MiB on all ranks)',
                who='Response', flush=True)

        # Perpare buffers for sending and receiving
        # counts - Number of elements to send to (s) or receive from (r) each rank
        # displs - Position of data to send to (s) or receive from (r) each rank
        sendbuf = np.zeros(my_sourcexsize * nnstride, dtype=self.dtype)
        recvbuf = np.zeros(my_targetxsize * nnstride, dtype=self.dtype)
        scounts_r = np.zeros(comm.size, dtype=int)
        sdispls_r = np.zeros(comm.size, dtype=int)
        rcounts_r = np.zeros(comm.size, dtype=int)
        rdispls_r = np.zeros(comm.size, dtype=int)

        displ = 0
        if comm.rank in target_indices_by_rank.keys():
            # This rank has some data to send. It will send to the ranks that are among the source keys
            sendbuf_by_rank = dict()
            for buf_nnx in self._iter_buffers():
                for destrank, xsize in source_xsize_by_rank.items():
                    size = xsize * nnstride
                    sendbuf_by_rank[destrank] = sendbuf[displ:displ+size]
                    scounts_r[destrank] = size
                    if size > 0:
                        sdispls_r[destrank] = displ
                    displ += size
                break
        displ = 0
        if comm.rank in source_indices_by_rank.keys():
            # This rank has some data to receive. It will receive from the ranks that are among the target keys
            recvbuf_by_rank = dict()
            for buf_nnx in self._iter_buffers():
                for destrank, xsize in target_xsize_by_rank.items():
                    size = xsize * nnstride
                    recvbuf_by_rank[destrank] = recvbuf[displ:displ+size]
                    rcounts_r[destrank] = size
                    if size > 0:
                        rdispls_r[destrank] = displ
                    displ += size
                break

        # Flattened nn-dimensions for splitting
        flatslices = [slice(start, start + nnstride, 1) for start in range(0, nnsize, nnstride)]
        grid = np.mgrid[:self.nnshape[0], :self.nnshape[1]]

        # Loop over real and imaginary parts and derivatives
        for (real, derivative), sendbuf_nnx, recvbuf_nnx in zip(
                self._iter_reim_derivatives(), self._iter_buffers(), target._iter_buffers()):

            # List of data to send and list of buffers where data should be received
            if comm.rank in target_indices_by_rank.keys():
                senddata_by_rank = {rank: sendbuf_nnx[self.nnellipsis + x_indices]
                                    for rank, x_indices in source_indices_by_rank.items()}
            if comm.rank in source_indices_by_rank.keys():
                recvdata_by_rank = {rank: recvbuf_nnx[self.nnellipsis + x_indices]
                                    for rank, x_indices in target_indices_by_rank.items()}
                # The target data may be smaller than what is given by the user
                recvdata_by_rank = {rank: recvdata_by_rank[rank][
                                        self.nnellipsis + tuple([slice(dim) for dim in xshape])]
                                    for rank, xshape in target_xshape_by_rank.items()}
                for data, xshape in zip(recvdata_by_rank.values(), target_xshape_by_rank.values()):
                    assert data.shape[2:] == xshape, str(data.shape[2:]) + ' != ' + str(xshape)

            # Loop over the data in splits
            for flatslice in flatslices:
                # Slices of nn
                nnslices = (grid[0].ravel()[flatslice], grid[1].ravel()[flatslice])

                # Copy data to the contiguous send buffer
                if comm.rank in target_indices_by_rank.keys():
                    for destrank, buf in sendbuf_by_rank.items():
                        data = senddata_by_rank[destrank][nnslices].ravel()
                        buf[:len(data)] = data

                # Send the data
                comm.alltoallv(sendbuf, scounts_r, sdispls_r,
                               recvbuf, rcounts_r, rdispls_r)

                # Copy data from the contiguous receive buffer
                if comm.rank in source_indices_by_rank.keys():
                    for destrank, buf in recvbuf_by_rank.items():
                        # Copy the first elements of the receive buffer to the data position
                        datashape = recvdata_by_rank[destrank][nnslices].shape
                        datalen = np.prod(datashape, dtype=int)
                        buf = buf[:datalen]
                        recvdata_by_rank[destrank][nnslices] = buf.reshape(datashape)



from __future__ import annotations

from .density_matrix import DensityMatrix
from .buffer import DensityMatrixBuffer
from .frequency import FrequencyDensityMatricesFromDisk, FrequencyDensityMatricesFromWaveFunctions
from .time import (TimeDensityMatricesFromWaveFunctions, ConvolutionDensityMatricesFromDisk,
                   ConvolutionDensityMatricesFromFrequency, ConvolutionDensityMatricesFromWaveFunctions)

__all__ = [
    'DensityMatrix',
    'DensityMatrixBuffer',
    'TimeDensityMatricesFromWaveFunctions',
    'ConvolutionDensityMatricesFromDisk',
    'ConvolutionDensityMatricesFromFrequency',
    'ConvolutionDensityMatricesFromWaveFunctions',
    'FrequencyDensityMatricesFromDisk',
    'FrequencyDensityMatricesFromWaveFunctions',
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar
import numpy as np
from numpy.typing import NDArray

from gpaw.mpi import world
from gpaw.io import Writer as GPAWWriter

from ..density_matrices.base import WorkMetadata, WorkMetadataT, BaseDensityMatrices
from ..density_matrices.frequency import FrequencyDensityMatrices, FrequencyDensityMatrixMetadata
from ..density_matrices.time import ConvolutionDensityMatrices, ConvolutionDensityMatrixMetadata
from ..calculators.base import BaseObservableCalculator
from ..voronoi import VoronoiWeights, EmptyVoronoiWeights, atom_projections_to_numpy
from ..utils import Result, ResultKeys


class ResultsCollector(ABC, Generic[WorkMetadataT]):

    """ Utility class to collect result arrays for different
    times, pulses, or frequencies.

    Parameters
    ----------
    resultkeys
        Result keys to be collected.
    additional_dimension
        Shape of additional dimension(s) due to the different times, frequencies, etc.
    additional_suffix
        String prepended to the suffix if each key.
    """

    def __init__(self,
                 calc: BaseObservableCalculator,
                 calc_kwargs: dict[str, Any],
                 resultkeys: ResultKeys,
                 additional_suffix: str,
                 additional_dimension: tuple[int, ...],
                 exclude: list[str] = []):
        self.calc = calc
        self.calc_kwargs = calc_kwargs
        self.resultkeys = resultkeys.__copy__()
        for key in exclude:
            if key in self.resultkeys:
                self.resultkeys.remove(key)
        self.additional_dimension = additional_dimension
        self.additional_suffix = additional_suffix

        # Create the new result keys for the aggregated data
        self.collect_resultkeys = ResultKeys()
        for key, shape, dtype in self.resultkeys:
            newkey = self.format_key(key)
            self.collect_resultkeys.add_key(newkey, additional_dimension + shape, dtype)

        self.result = Result(mutable=True)

    def empty_results(self):
        if world.rank == 0:
            self.result.create_all_zeros(self.collect_resultkeys)

    def finalize_results(self):
        pass

    def format_key(self,
                   key: str) -> str:
        """ Add the new suffix to the key.

        Parameters
        ----------
        key
            Original result key.

        Returns
        -------
        New result key with the added suffix.
        """
        shape, _ = self.resultkeys[key]
        if len(shape) == 0:
            return key + f'_{self.additional_suffix}'

        s = key.split('_')
        assert len(s) > 1
        s[-1] = self.additional_suffix + s[-1]
        return '_'.join(s)

    @abstractmethod
    def accumulate_results(self,
                           work: WorkMetadataT,
                           result: Result):
        pass


ResultsCollectorT = TypeVar('ResultsCollectorT', bound=ResultsCollector)


class TimeResultsCollector(ResultsCollector):

    """ Collect results after convolution with different pulses.

    The letter t is prepended to the suffix of the result keys to indicate
    an additional dimension of time.

    Parameters
    ----------
    calc
        Calculator.
    calc_kwargs
        Keyword arguments passed to the icalculate function.
    exclude
        Keys that are excluded from collection.
    """

    def __init__(self,
                 calc: BaseObservableCalculator,
                 calc_kwargs: dict[str, Any],
                 exclude: list[str] = []):
        assert isinstance(calc.density_matrices, ConvolutionDensityMatrices)
        assert len(calc.density_matrices.pulses) == 1
        Nt = len(calc.times)

        resultkeys = calc.get_result_keys(**calc_kwargs)
        super().__init__(calc, calc_kwargs, resultkeys,
                         additional_suffix='t', additional_dimension=(Nt, ), exclude=exclude)

    def accumulate_results(self,
                           work: ConvolutionDensityMatrixMetadata,
                           result: Result):
        assert isinstance(work, ConvolutionDensityMatrixMetadata)
        assert world.rank == 0

        for key, _, _ in self.resultkeys:
            newkey = self.format_key(key)
            self.result.set_to(newkey, work.globalt, result[key])


class TimeAverageResultsCollector(ResultsCollector):

    """ Collect results and average over times.

    Parameters
    ----------
    calc
        Calculator.
    calc_kwargs
        Keyword arguments passed to the icalculate function.
    exclude
        Keys that are excluded from collection.
    """

    def __init__(self,
                 calc: BaseObservableCalculator,
                 calc_kwargs: dict[str, Any],
                 exclude: list[str] = []):
        assert isinstance(calc.density_matrices, ConvolutionDensityMatrices)
        assert len(calc.density_matrices.pulses) == 1

        resultkeys = calc.get_result_keys(**calc_kwargs)
        super().__init__(calc, calc_kwargs, resultkeys,
                         additional_suffix='', additional_dimension=(), exclude=exclude)

    def accumulate_results(self,
                           work: ConvolutionDensityMatrixMetadata,
                           result: Result):
        assert isinstance(work, ConvolutionDensityMatrixMetadata)
        assert world.rank == 0

        for key, _, _ in self.resultkeys:
            newkey = self.format_key(key)
            self.result.add_to(newkey, slice(None), result[key])

    def finalize_results(self):
        if world.rank > 0:
            return

        nt = len(self.calc.density_matrices.times)
        for key, _, _ in self.collect_resultkeys:
            self.result[key] /= nt


class PulseConvolutionResultsCollector(ResultsCollector):

    """ Collect results after convolution with different pulses.

    The letters pt are prepended to the suffix of the result keys to indicate
    an additional dimension of pulse and time.

    Parameters
    ----------
    calc
        Calculator.
    calc_kwargs
        Keyword arguments passed to the icalculate function.
    exclude
        Keys that are excluded from collection.
    """

    def __init__(self,
                 calc: BaseObservableCalculator,
                 calc_kwargs: dict[str, Any],
                 exclude: list[str] = []):
        assert isinstance(calc.density_matrices, ConvolutionDensityMatrices)
        Np = len(calc.pulses)
        Nt = len(calc.times)

        resultkeys = calc.get_result_keys(**calc_kwargs)
        super().__init__(calc, calc_kwargs, resultkeys,
                         additional_suffix='pt', additional_dimension=(Np, Nt), exclude=exclude)

    def accumulate_results(self,
                           work: ConvolutionDensityMatrixMetadata,
                           result: Result):
        assert isinstance(work, ConvolutionDensityMatrixMetadata)
        assert world.rank == 0

        for key, _, _ in self.resultkeys:
            newkey = self.format_key(key)
            self.result.set_to(newkey, (work.globalp, work.globalt), result[key])


class PulseConvolutionAverageResultsCollector(ResultsCollector):

    """ Collect results after convolution with different pulses, average over times.

    The letter p is prepended to the suffix of the result keys to indicate
    an additional dimension of pulse.

    Parameters
    ----------
    calc
        Calculator.
    calc_kwargs
        Keyword arguments passed to the icalculate function.
    exclude
        Keys that are excluded from collection.
    """

    def __init__(self,
                 calc: BaseObservableCalculator,
                 calc_kwargs: dict[str, Any],
                 exclude: list[str] = []):
        assert isinstance(calc.density_matrices, ConvolutionDensityMatrices)
        Np = len(calc.pulses)

        resultkeys = calc.get_result_keys(**calc_kwargs)
        super().__init__(calc, calc_kwargs, resultkeys,
                         additional_suffix='p', additional_dimension=(Np, ), exclude=exclude)

    def accumulate_results(self,
                           work: ConvolutionDensityMatrixMetadata,
                           result: Result):
        assert isinstance(work, ConvolutionDensityMatrixMetadata)
        assert world.rank == 0

        for key, _, _ in self.resultkeys:
            newkey = self.format_key(key)
            self.result.add_to(newkey, work.globalp, result[key])

    def finalize_results(self):
        if world.rank > 0:
            return

        nt = len(self.calc.density_matrices.times)
        for key, _, _ in self.collect_resultkeys:
            self.result[key] /= nt


class FrequencyResultsCollector(ResultsCollector):

    """ Collect results in the frequency domain.

    This class should work with the Fourier transform of
    the real part of density matrices.

    The letter w is prepended to the suffix of the result keys to indicate
    an additional dimension of frequency.

    Parameters
    ----------
    calc
        Calculator.
    calc_kwargs
        Keyword arguments passed to the icalculate function.
    exclude
        Keys that are excluded from collection
    """

    def __init__(self,
                 calc: BaseObservableCalculator,
                 calc_kwargs: dict[str, Any],
                 exclude: list[str] = []):
        assert isinstance(calc.density_matrices, FrequencyDensityMatrices)
        Nw = len(calc.frequencies)
        assert 'Im' not in calc.density_matrices.reim

        resultkeys = calc.get_result_keys(**calc_kwargs)
        super().__init__(calc, calc_kwargs, resultkeys,
                         additional_suffix='w', additional_dimension=(Nw, ), exclude=exclude)

    def accumulate_results(self,
                           work: FrequencyDensityMatrixMetadata,
                           result: Result):
        assert isinstance(work, FrequencyDensityMatrixMetadata)
        assert world.rank == 0

        for key, _, _ in self.resultkeys:
            newkey = self.format_key(key)
            self.result.set_to(newkey, work.globalw, result[key])


class Writer(Generic[ResultsCollectorT]):

    def __init__(self, collector: ResultsCollectorT):
        self._collector = collector
        self._ulm_tag = 'RhodentResults'

    @property
    def collector(self) -> ResultsCollectorT:
        return self._collector

    @property
    def calc(self) -> BaseObservableCalculator:
        return self.collector.calc

    @property
    def density_matrices(self) -> BaseDensityMatrices:
        return self.collector.calc.density_matrices

    @property
    def voronoi(self) -> VoronoiWeights:
        voronoi = self.calc.voronoi
        if voronoi is None:
            return EmptyVoronoiWeights()
        return voronoi

    @property
    def common_arrays(self) -> dict[str, NDArray[np.float64] | NDArray[np.int64] | int | float]:
        """ Dictionary of eigenvalues and limits. """
        imin, imax, amin, amax = self.calc.ksd.ialims()
        arrays: dict[str, NDArray[np.float64] | NDArray[np.int64] | int | float] = dict()
        arrays['eig_n'] = self.calc.eig_n
        arrays['eig_i'] = self.calc.eig_i
        arrays['eig_a'] = self.calc.eig_a
        arrays['imin'] = imin
        arrays['imax'] = imax
        arrays['amin'] = amin
        arrays['amax'] = amax

        return arrays

    @property
    def icalculate_kwargs(self) -> dict:
        """ Keyword arguments to icalculate. """
        return self.collector.calc_kwargs

    def fill_ulm(self,
                 writer,
                 work: WorkMetadata,
                 result: Result):
        """ Fill one entry of the ULM file.

        Parameters
        ----------
        writer
            Open ULM writer object.
        work
            Metadata to current piece of data.
        result
            Result containing the current observables.
        """
        raise NotImplementedError

    def write_empty_arrays_ulm(self, writer):
        """ Add empty arrays in to the ULM file.

        Parameters
        ----------
        writer
            Open ULM writer object.
        """
        raise NotImplementedError

    def calculate_data(self) -> Result:
        """ Calculate results on all ranks and return Result object.

        Returns
        -------
        Retult object. Is empty on non-root ranks.
        """
        self.collector.empty_results()

        for work, res in self.calc.icalculate_gather_on_root(**self.icalculate_kwargs):
            self.collector.accumulate_results(work, res)

        self.collector.finalize_results()

        return self.collector.result

    def calculate_and_save_npz(self,
                               out_fname: str,
                               write_extra: dict[str, Any] = dict()):
        """ Calculate results on all ranks and save to npz file.

        Parameters
        ----------
        out_fname
            File name.
        """
        result = self.calculate_data()

        if world.rank > 0:
            return

        atom_projections = atom_projections_to_numpy(self.voronoi.atom_projections)
        np.savez(out_fname, **self.common_arrays, **result._data,  # type: ignore
                 atom_projections=atom_projections)
        self.calc.log_parallel(f'Written {out_fname}', flush=True)

    def calculate_and_save_ulm(self,
                               out_fname: str,
                               write_extra: dict[str, Any] = dict()):
        """ Calculate results on all ranks and save to ULM file.

        Parameters
        ----------
        out_fname
            File name.
        """
        self.collector.empty_results()

        with GPAWWriter(out_fname, world, mode='w', tag=self._ulm_tag[:16]) as writer:
            writer.write(version=1)
            writer.write('atom_projections', self.voronoi.atom_projections)
            writer.write(**(self.common_arrays if world.rank == 0 else dict()))

            self.write_empty_arrays_ulm(writer)

            for work, res in self.calc.icalculate_gather_on_root(**self.icalculate_kwargs):
                self.fill_ulm(writer, work, res)
                self.collector.accumulate_results(work, res)

            self.collector.finalize_results()
            writer.write(**self.collector.result._data)

        if world.rank == 0:
            self.calc.log_parallel(f'Written {out_fname}', flush=True)



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from gpaw.tddft.units import au_to_eV

from ..density_matrices.base import WorkMetadata
from ..density_matrices.time import ConvolutionDensityMatrices
from .writer import Writer, ResultsCollector
from ..utils import Result, get_gaussian_pulse_values


class EnergyWriter(Writer):

    """ Calculate energy contributions

    Parameters
    ----------
    collector
        ResultsCollector object
    """

    def __init__(self,
                 collector: ResultsCollector,
                 only_one_pulse: bool):
        super().__init__(collector)
        self.only_one_pulse = only_one_pulse
        self._ulm_tag = 'EnergyDecomposition'
        if only_one_pulse:
            if isinstance(self.density_matrices, ConvolutionDensityMatrices):
                assert len(self.density_matrices.pulses) == 1, 'Only one pulse allowed'
        else:
            assert isinstance(self.density_matrices, ConvolutionDensityMatrices)

    @property
    def common_arrays(self) -> dict[str, NDArray[np.float64] | NDArray[np.int64] | int | float]:
        common = super().common_arrays

        if self.calc.sigma is not None:
            # There is an energy grid
            common['sigma'] = self.calc.sigma
            common['energy_o'] = np.array(self.calc.energies_occ)
            common['energy_u'] = np.array(self.calc.energies_unocc)

        assert isinstance(self.density_matrices, ConvolutionDensityMatrices)
        common['time_t'] = self.density_matrices.times * 1e-3

        # If pulses are Gaussian pulses, then get dictionaries of 'pulsefreq' and 'pulsefwhm'
        pulsedicts = [get_gaussian_pulse_values(pulse) for pulse in self.density_matrices.pulses]

        try:
            pulsefreqs = [d['pulsefreq'] for d in pulsedicts]
            pulsefwhms = [d['pulsefwhm'] for d in pulsedicts]

            if self.only_one_pulse:
                common['pulsefreq'] = pulsefreqs[0]
                common['pulsefwhm'] = pulsefwhms[0]
            else:
                common['pulsefreq_p'] = np.array(pulsefreqs)
                common['pulsefwhm_p'] = np.array(pulsefwhms)
        except KeyError:
            # Not GaussianPulses
            pass

        common['resonant_low'] = self.calc._filter_pair_low * au_to_eV  # type: ignore
        common['resonant_high'] = self.calc._filter_pair_high * au_to_eV  # type: ignore
        return common

    def fill_ulm(self,
                 writer,
                 work: WorkMetadata,
                 result: Result):
        assert self.only_one_pulse
        if self.collector.calc_kwargs['yield_total_E_ou']:
            writer.fill(result['E_ou'])

    def write_empty_arrays_ulm(self, writer):
        assert self.only_one_pulse
        if not self.collector.calc_kwargs['yield_total_E_ou']:
            return
        shape_ou = (len(self.calc.energies_occ), len(self.calc.energies_unocc))
        Nt = len(self.density_matrices.times)
        writer.add_array('E_tou', (Nt, ) + shape_ou, dtype=float)



from __future__ import annotations

from typing import Collection

import numpy as np
from pathlib import Path

from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition

from .base import BaseResponse
from ..perturbation import create_perturbation, PerturbationLike, NoPerturbation
from ..density_matrices.frequency import (FrequencyDensityMatrices,
                                          FrequencyDensityMatricesFromWaveFunctions)
from ..density_matrices.time import (ConvolutionDensityMatrices,
                                     ConvolutionDensityMatricesFromWaveFunctions,
                                     TimeDensityMatricesFromWaveFunctions)
from ..utils import Logger
from ..typing import Array1D


class ResponseFromWaveFunctions(BaseResponse):

    """ Response from time-dependent wave functions file written by GPAW.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object or file name.
    wfs_fname
        File name of the time-dependent wave functions file, written by :class:`WaveFunctionsWriter`.
    perturbation
        Perturbation that was present during time propagation.
    calc_size
        Size of the calculation communicator.
    """

    def __init__(self,
                 wfs_fname: Path | str,  # File name of wfs.ulm file
                 ksd: KohnShamDecomposition | str,
                 perturbation: PerturbationLike = None,
                 calc_size: int = 1,
                 stridet: int = 1):
        super().__init__(ksd=ksd,
                         perturbation=perturbation,
                         calc_size=calc_size)

        self.wfs_fname = str(wfs_fname)

        # Options for reading the wfs.ulm file
        self.stridet = stridet

    def __str__(self) -> str:
        lines = [f'{self.__class__.__name__}']
        lines += [f'  ksd: {self.ksd.filename if self.ksd.filename is not None else "From calc"}']
        lines += [f'  wfs_fname: {self.wfs_fname}']
        lines += ['  perturbation:']
        lines += ['    ' + line for line in str(self.perturbation).split('\n')]
        return '\n'.join(lines)

    def _get_time_density_matrices(self,
                                   times: list[float] | Array1D[np.float64],
                                   pulses: Collection[PerturbationLike],
                                   derivative_order_s: list[int] = [0],
                                   real: bool = True,
                                   imag: bool = True,
                                   log: Logger | None = None,
                                   ) -> ConvolutionDensityMatrices:
        density_matrices: ConvolutionDensityMatrices

        # Perform convolution if pulses differ from perturbation or higher derivatives needed
        if (all(create_perturbation(pulse) == self.perturbation for pulse in pulses) and
                all(derivative == 0 for derivative in derivative_order_s)):
            # No convolution needed
            density_matrices = TimeDensityMatricesFromWaveFunctions(
                ksd=self.ksd,
                wfs_fname=self.wfs_fname,
                times=times,
                real=real,
                imag=imag,
                log=log,
                calc_size=self.calc_size,
                stridet=self.stridet)
        else:
            # Perform convolution
            if isinstance(self.perturbation, NoPerturbation):
                raise ValueError('Perturbation must be given to perform pulse convolution.')
            density_matrices = ConvolutionDensityMatricesFromWaveFunctions(
                ksd=self.ksd,
                wfs_fname=self.wfs_fname,
                perturbation=self.perturbation,
                pulses=pulses,
                times=times,
                derivative_order_s=derivative_order_s,
                real=real,
                imag=imag,
                log=log,
                calc_size=self.calc_size,
                stridet=self.stridet)

        return density_matrices

    def _get_frequency_density_matrices(self,
                                        frequencies: list[float] | Array1D[np.float64],
                                        frequency_broadening: float = 0,
                                        real: bool = True,
                                        imag: bool = True,
                                        log: Logger | None = None,
                                        ) -> FrequencyDensityMatrices:
        if isinstance(self.perturbation, NoPerturbation):
            raise ValueError('Perturbation must be given to normalize Fourier transform.')
        density_matrices = FrequencyDensityMatricesFromWaveFunctions(
            ksd=self.ksd,
            wfs_fname=self.wfs_fname,
            perturbation=self.perturbation,
            frequencies=frequencies,
            frequency_broadening=frequency_broadening,
            real=real,
            imag=imag,
            calc_size=self.calc_size,
            log=log,
            stridet=self.stridet)

        return density_matrices



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray, DTypeLike

from typing import Iterator


class ResultKeys():

    """ List of result keys.

    """

    _keys_dimensions_dtypes: dict[str, tuple[tuple[int, ...], np.dtype]]

    def __init__(self,
                 *scalar_keys):
        self._keys_dimensions_dtypes = dict()

        for key in scalar_keys:
            self.add_key(key, (), float)

    def add_key(self,
                key: str,
                shape: tuple[int, ...] | int = (),
                dtype: DTypeLike = float):
        """ Add a new result key.

        Parameters
        ----------
        key
            Name of result.
        shape
            Shape of result (at one time or frequency instance). Default is scalar.
        dtype
            Result dtype.
        """

        assert isinstance(key, str)
        if isinstance(shape, int):
            shape = (shape, )
        assert all([isinstance(d, int) for d in shape])
        dtype = np.dtype(dtype)
        self._keys_dimensions_dtypes[key] = (shape, dtype)

    def remove(self,
               key: str):
        assert key in self
        self._keys_dimensions_dtypes.pop(key)

    def __contains__(self,
                     key: str) -> bool:
        return key in self._keys_dimensions_dtypes.keys()

    def __iter__(self) -> Iterator[tuple[str, tuple[int, ...], np.dtype]]:
        for key, (shape, dtype) in self._keys_dimensions_dtypes.items():
            yield key, shape, dtype

    def __getitem__(self,
                    key: str) -> tuple[tuple[int, ...], np.typing.DTypeLike]:
        assert key in self._keys_dimensions_dtypes, f'Key {key} not among keys'
        return self._keys_dimensions_dtypes[key]

    def __copy__(self):
        cpy = ResultKeys()
        cpy._keys_dimensions_dtypes.update(self._keys_dimensions_dtypes)
        return cpy


class Result:

    """ Class holding results.

    """

    _data: dict[str, NDArray[np.float64]]

    def __init__(self,
                 mutable: bool = False):
        self._data = dict()
        self._mutable = mutable

    def __contains__(self,
                     key: str) -> bool:
        return key in self._data

    def __setitem__(self,
                    key: str,
                    value: np.typing.ArrayLike | int):
        if not self._mutable:
            assert key not in self._data, f'Key {key} is already among results'
        if np.ndim(value) == 0:
            value = np.array([value])
        self._data[key] = np.ascontiguousarray(value)

    def __getitem__(self,
                    key: str) -> NDArray[np.float64]:
        assert key in self._data, f'Key {key} not among results'
        return self._data[key]

    def __str__(self) -> str:
        lines = [f'{self.__class__.__name__} with arrays (dimensions)']

        for key, data in self._data.items():
            lines.append(f'  {key} {data.shape}')

        return '\n'.join(lines)

    def set_to(self,
               key: str,
               idx,
               value: np.typing.ArrayLike | int | float):
        if np.ndim(self._data[key][idx]) == 0:
            assert np.size(value) == 1
            value = np.atleast_1d(value)[0]
        self._data[key][idx] = value

    def add_to(self,
               key: str,
               idx,
               value: np.typing.ArrayLike | int | float):
        if np.ndim(self._data[key][idx]) == 0:
            assert np.size(value) == 1
            value = np.atleast_1d(value)[0]
        self._data[key][idx] += value

    def create_all_empty(self,
                         keys: ResultKeys):
        for key, shape, dtype in keys:
            if key in self:
                continue
            self[key] = np.empty(shape, dtype=dtype)

    def create_all_zeros(self,
                         keys: ResultKeys):
        for key, shape, dtype in keys:
            if key in self:
                continue
            self[key] = np.zeros(shape, dtype=dtype)

    def remove(self,
               key: str):
        assert key in self._data
        self._data.pop(key)

    def empty(self,
              key: str,
              keys: ResultKeys):
        shape, dtype = keys[key]
        self[key] = np.empty(shape, dtype=dtype)

    def assert_keys(self,
                    keys: ResultKeys):
        copy = dict(self._data)
        try:
            for key, shape, dtype in keys:
                array = copy.pop(key)
                if len(shape) == 0:
                    assert array.shape == (1, ), f'{array.shape} != (1,)'
                else:
                    assert array.shape == shape, f'{array.shape} != {shape}'
                assert array.dtype == dtype, f'{array.dtype} != {dtype}'
        except KeyError:
            raise AssertionError(f'Key {key} missing from Result')
        assert len(copy) == 0, f'Result has additional keys {copy.keys()}'

    def send(self,
             keys: ResultKeys,
             rank,
             comm):
        self.assert_keys(keys)
        for vi, (key, _, _) in enumerate(keys):
            value = self._data[key]
            comm.send(value, rank, tag=100 + vi)

    def inplace_receive(self,
                        keys: ResultKeys,
                        rank: int,
                        comm):
        self.assert_keys(keys)
        for vi, (key, _, _) in enumerate(keys):
            value = self._data[key]
            comm.receive(value, rank, tag=100 + vi)



from __future__ import annotations


from os import getenv


def rhodent_getenv(variable) -> str:
    """ Get value of environment variable, or default.

    Possible environment variables are:

    - `RHODENT_RESPONSE_MAX_MEM` - Value in units of MiB. When constructing
      response, try to limit memory usage below this value.
    - `RHODENT_RESPONSE_MAX_MEM_PER_RANK` - Value in units of MiB.
      Ignored if `RHODENT_RESPONSE_MAX_MEM` is set.
    - `RHODENT_REDISTRIBUTE_MAXSIZE` - Maximal number of elements in single redistribute call.
      Larger arrays are split.

    Parameters
    ----------
    variable
        Name of variable, without the ``'RHODENT_'`` prefix.

    Returns
    -------
    Value of variable as string.
    """

    defaults = {'REDISTRIBUTE_MAXSIZE': '1e7',
                'RESPONSE_MAX_MEM': '',
                'RESPONSE_MAX_MEM_PER_RANK': '1000',
                }

    if variable not in defaults.keys():
        allowed = '\n'.join([f'  RHODENT_{var}' for var in defaults.keys()])
        raise ValueError(f'Environment variable RHODENT_{variable} is unknown. '
                         f'Allowed variables are:\n{allowed}')

    return getenv(f'RHODENT_{variable}', defaults[variable])


def get_float(variable) -> float:
    strvalue = rhodent_getenv(variable)
    try:
        value = float(strvalue)
    except ValueError:
        raise ValueError(f'Expected environment variable RHODENT_{variable} to '
                         f'be castable to float. Value is {strvalue!r}.')

    return value


def get_response_max_mem(comm_size: int):
    if rhodent_getenv('RESPONSE_MAX_MEM') == '':
        return get_float('RESPONSE_MAX_MEM_PER_RANK') * comm_size

    return get_float('RESPONSE_MAX_MEM')



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from gpaw.mpi import broadcast, world, SerialCommunicator
from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition

from ..typing import ArrayIsOnRootRank, DistributedArray, Communicator


class DensityMatrix:

    """ Wrapper for the density matrix in the Kohn-Sham basis at one moment
    in time or at one frequency.

    The plain density matrix and/or derivatives thereof may be stored.

    Parameters
    ----------
    ksd
        KohnShamDecomposition object.
    matrices
        Dictionary mapping derivative orders (0, 1, 2) for zeroth,
        first, second derivative, .. to arrays storing the matrices.
    comm
        MPI communicator. Serial communicator by default.
    """

    def __init__(self,
                 ksd: KohnShamDecomposition,
                 matrices: dict[int, NDArray[np.complex128] | None],
                 comm: Communicator | None = None):
        self._ksd = ksd
        if comm is None:
            comm = SerialCommunicator()  # type: ignore
        self._comm = comm

        # Calculate occupation number difference
        f_n = self.ksd.occ_un[0]
        imin, imax, amin, amax = self.ksd.ialims()
        self._f_ia = f_n[imin:imax+1, None] - f_n[None, amin:amax+1]
        self._f_ia[self._f_ia < 0] = 0

        # Calculate mask
        min_occdiff = min(self.ksd.f_p)
        mask_ia = self.f_ia >= min_occdiff

        self._matrices: dict[int, DistributedArray] = dict()
        self.derivative_desc = {0: 'Plain DM', 1: '1st DM derivative', 2: '2nd DM derivative'}

        # Save the matrices
        for derivative, rho in matrices.items():
            assert isinstance(derivative, int)
            if self.rank == 0:
                assert isinstance(rho, np.ndarray), rho
                self._matrices[derivative] = rho * mask_ia
            else:
                assert rho is None
                self._matrices[derivative] = ArrayIsOnRootRank()

    @property
    def ksd(self) -> KohnShamDecomposition:
        """ Kohn-Sham decomposition object. """
        return self._ksd

    @property
    def rank(self) -> int:
        """ MPI rank of the communicator. """
        return self.comm.rank

    @property
    def comm(self) -> Communicator:
        return self._comm  # type: ignore

    @property
    def f_ia(self) -> DistributedArray:
        """ Occupation number difference :math:`f_{ia}`. """
        return self._f_ia

    @property
    def rho_ia(self) -> DistributedArray:
        r""" Electron-hole part of induced density matrix :math:`\delta rho_{ia}`. """
        try:
            return self._matrices[0]
        except KeyError:
            raise ValueError('Plain density matrix not in {self._matrices.keys()}')

    @property
    def drho_ia(self) -> DistributedArray:
        r""" First time derivative of :math:`\delta rho_{ia}`. """
        try:
            return self._matrices[1]
        except KeyError:
            raise ValueError('First derivative of density matrix not in {self._matrices.keys()}')

    @property
    def ddrho_ia(self) -> DistributedArray:
        r""" Second time derivative of :math:`\delta rho_{ia}`. """
        try:
            return self._matrices[2]
        except KeyError:
            raise ValueError('Second derivative of density matrix not in {self._matrices.keys()}')

    @property
    def Q_ia(self) -> DistributedArray:
        r""" The quantity

        .. math::
            Q_{ia} = \frac{2 \mathrm{Re}\:\delta\rho_{ia}}{\sqrt{2 f_{ia}}}

        where :math:`f_{ia}` is the occupation number difference of pair :math:`ia`.
        """
        return self._divide_by_sqrt_fia(np.sqrt(2) * self.rho_ia.real)

    @property
    def P_ia(self) -> DistributedArray:
        r""" The quantity

        .. math::
            P_{ia} = \frac{2 \mathrm{Im}\:\delta\rho_{ia}}{\sqrt{2 f_{ia}}}

        where :math:`f_{ia}` is the occupation number difference of pair :math:`ia`.
        """
        return self._divide_by_sqrt_fia(np.sqrt(2) * self.rho_ia.imag)

    @property
    def dQ_ia(self) -> DistributedArray:
        r""" First time derivative of :math:`Q_{ia}`. """
        return self._divide_by_sqrt_fia(np.sqrt(2) * self.drho_ia.real)

    @property
    def dP_ia(self) -> DistributedArray:
        r""" First time derivative of :math:`P_{ia}`. """
        return self._divide_by_sqrt_fia(np.sqrt(2) * self.drho_ia.imag)

    @property
    def ddQ_ia(self) -> DistributedArray:
        r""" Second time derivative of :math:`Q_{ia}`. """
        return self._divide_by_sqrt_fia(np.sqrt(2) * self.ddrho_ia.real)

    @property
    def ddP_ia(self) -> DistributedArray:
        r""" Second time derivative of :math:`P_{ia}`. """
        return self._divide_by_sqrt_fia(np.sqrt(2) * self.ddrho_ia.imag)

    def _divide_by_sqrt_fia(self,
                            X_ia: DistributedArray) -> DistributedArray:
        r""" Divide by :math:`\sqrt{f_{ia}}` where :math:`f_{ia} \neq 0`.
        Leave everything else as 0."""
        if self.rank > 0:
            assert isinstance(X_ia, ArrayIsOnRootRank)
            return ArrayIsOnRootRank()
        assert not isinstance(X_ia, ArrayIsOnRootRank)
        flt_ia = self.f_ia != 0
        Y_ia = np.zeros_like(X_ia)
        Y_ia[flt_ia] = X_ia[flt_ia] / np.sqrt(self.f_ia[flt_ia])

        return Y_ia

    def copy(self) -> DensityMatrix:
        """ Copy the density matrix. """
        matrices: dict[int, NDArray[np.complex128] | None] = {
            derivative: np.array(matrix) for derivative, matrix in self._matrices.items()}
        dm = DensityMatrix(ksd=self.ksd, matrices=matrices, comm=self.comm)
        return dm

    @classmethod
    def broadcast(cls,
                  density_matrix: DensityMatrix | None,
                  ksd: KohnShamDecomposition,
                  root: int,
                  dm_comm,
                  comm) -> DensityMatrix:
        """ Broadcast a density matrix object which is on one rank to all other ranks.

        Parameters
        ----------
        density_matrix
            The density matrix to be broadcast on the root rank, and ``None`` on other ranks.
        ksd
            KohnShamDecomposition object.
        root
            Rank of the process that has the original data.
        dm_comm
            Must be identical to communicator of :attr:`density_matrix`.
        comm
            MPI communicator. Must be complementary to the communicator of :attr:`density_matrix`.
        """
        matrices: dict[int, NDArray[np.complex128] | None]
        # Broadcast necessary metadata
        if comm.rank == root:
            assert density_matrix is not None
            matrix_shapes_dtypes = {derivative: (matrix.shape, matrix.dtype)
                                    for derivative, matrix in density_matrix._matrices.items()}
            broadcast(matrix_shapes_dtypes, root=root, comm=comm)
            matrices = {derivative: None if isinstance(arr, ArrayIsOnRootRank) else arr
                        for derivative, arr in density_matrix._matrices.items()}
        else:
            assert density_matrix is None
            matrix_shapes_dtypes = broadcast(None, root=root, comm=comm)

        if comm.rank != root:
            if dm_comm.rank == 0:
                matrices = {derivative: np.empty(shape, dtype=dtype)
                            for derivative, (shape, dtype) in matrix_shapes_dtypes.items()}
            else:
                matrices = {derivative: None
                            for derivative, (shape, dtype) in matrix_shapes_dtypes.items()}

        if dm_comm.size > 1 and comm.size > 1:
            # Make sure communicators are complementary
            comm_members = comm.get_members()
            dm_members = dm_comm.get_members()

            intersect = set(comm_members) & set(dm_members)
            intersect.remove(world.rank)
            assert len(intersect) == 0, f'{comm_members} / {dm_members}'

        # On density matrix non-root ranks, return ArrayIsOnRootRank()
        if dm_comm.rank > 0:
            return DensityMatrix(ksd=ksd, matrices=matrices, comm=dm_comm)

        # Broadcast the matrices
        for derivative, matrix in matrices.items():
            comm.broadcast(np.ascontiguousarray(matrix), root)

        if comm.rank == root:
            assert density_matrix is not None
            return density_matrix
        else:
            return DensityMatrix(ksd=ksd, matrices=matrices, comm=dm_comm)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Union
from numbers import Number
import numpy as np
from numpy.typing import NDArray

from gpaw.lcaotddft.laser import Laser, create_laser


def create_perturbation(perturbation: PerturbationLike):
    if isinstance(perturbation, Perturbation):
        return perturbation
    if perturbation is None:
        return NoPerturbation()
    if isinstance(perturbation, Laser):
        return PulsePerturbation(perturbation)

    assert isinstance(perturbation, dict)
    if perturbation['name'] == 'none':
        return NoPerturbation
    if perturbation['name'] == 'deltakick':
        return DeltaKick(strength=perturbation['strength'])
    pulse = create_laser(perturbation)
    return PulsePerturbation(pulse)


class Perturbation(ABC):

    """ Perturbation. """

    def timestep(self,
                 times: NDArray[np.float64]):
        if len(times) < 2:
            raise ValueError('At least two times must be given to get a timestep.')
        dt = times[1] - times[0]
        if not np.allclose(times[1:] - dt, times[:-1]):
            raise ValueError('The time step may not vary.')
        return dt

    def frequencies(self,
                    times: NDArray[np.float64],
                    padnt: int | None = None) -> NDArray[np.float64]:
        """ Get the frequencies grid.

        Parameters
        ----------
        times
            Time grid in atomic units.
        padnt
            Length of data, including zero padding. Default is not zero padding.

        Returns
        -------
        Frequencies grid in atomic units.
        """
        timestep = self.timestep(times)
        if padnt is None:
            padnt = len(times)
        return 2 * np.pi * np.fft.rfftfreq(padnt, timestep)

    @abstractmethod
    def normalize_frequency_response(self,
                                     data: NDArray[np.float64],
                                     times: NDArray[np.float64],
                                     padnt: int,
                                     axis: int = -1) -> NDArray[np.complex128]:
        """
        Calculate a normalized response in the frequency domain, i.e., the
        response to a unity strength delta kick. For example, polarizability.

        Parameters
        ----------
        data
            Real valued response in the time domain to this perturbation.
        times
            Time grid in atomic units.
        axis
            Axis corresponding to time dimension.
        padnt
            Length of data, including zero padding.

        Returns
        -------
        Normalized response in the frequency domain.
        """
        raise NotImplementedError

    def normalize_time_response(self,
                                data: NDArray[np.float64],
                                times: NDArray[np.float64],
                                axis: int = -1) -> NDArray[np.float64]:
        """
        Transform response in the time domain to a "normalized response",
        which is the response to a unity strength delta kick.

        Parameters
        ----------
        data
            Real valued response in the time domain to this perturbation.
        times
            Time grid in atomic units.
        axis
            Axis corresponding to time dimension.

        Returns
        -------
        Normalized response in the time domain.
        """
        from .utils import fast_pad

        dt = times[1] - times[0]
        nt = len(times)
        padnt = fast_pad(nt)

        data = data.swapaxes(axis, -1)  # Put the time dimension last

        # Calculate the normalized response in the frequency domain
        data_w = self.normalize_frequency_response(data, times, padnt, axis=-1)

        # Fourier transform back to time tomain
        data_t = np.fft.irfft(data_w, n=padnt, axis=-1)[..., :nt] / dt

        data_t = data_t.swapaxes(axis, -1)
        return data_t

    @abstractmethod
    def amplitude(self,
                  times: NDArray[np.float64]) -> NDArray[np.float64]:
        """
        Perturbation amplitudes in time domain.

        Parameters
        ----------
        times
            Time grid in atomic units.

        Returns
        -------
        Perturbation at the given times.
        """
        raise NotImplementedError

    @abstractmethod
    def fourier(self,
                times: NDArray[np.float64],
                padnt: int | None = None) -> NDArray[np.complex128]:
        """
        Fourier transform of perturbation.

        Parameters
        ----------
        times
            Time grid in atomic units.
        padnt
            Length of data, including zero padding. Default is no added zero padding.

        Returns
        -------
        Fourier transform of the perturbation at the frequency grid \
        given by :func:`frequencies`.
        """
        raise NotImplementedError

    @abstractmethod
    def todict(self) -> dict[str, Any]:
        raise NotImplementedError

    def __eq__(self, other) -> bool:
        """ Equal if dicts are identical (up to numerical tolerance).
        """
        try:
            d1 = self.todict()
            d2 = other.todict()
        except AttributeError:
            return False

        if d1.keys() != d2.keys():
            return False

        for key in d1.keys():
            if isinstance(d1[key], Number) and isinstance(d2[key], Number):
                if not np.isclose(d1[key], d2[key]):
                    return False
            else:
                if not d1[key] == d2[key]:
                    return False

        return True


PerturbationLike = Union[Perturbation, Laser, dict, None]


class NoPerturbation(Perturbation):

    """ No perturbation

    Used to indicate that we do not know the perturbation,
    and that it should not matter.
    """

    def __init__(self):
        pass

    def amplitude(self,
                  times: NDArray[np.float64]) -> NDArray[np.float64]:
        raise RuntimeError('Not possible for no perturbation.')

    def fourier(self,
                times: NDArray[np.float64],
                padnt: int | None = None) -> NDArray[np.complex128]:
        raise RuntimeError('Not possible for no perturbation')

    def normalize_frequency_response(self,
                                     data: NDArray[np.float64],
                                     times: NDArray[np.float64],
                                     padnt: int,
                                     axis: int = -1) -> NDArray[np.complex128]:
        raise RuntimeError('Not possible for no perturbation')

    def __str__(self) -> str:
        return 'No perturbation'

    def todict(self) -> dict[str, Any]:
        return {'name': 'none'}


class DeltaKick(Perturbation):

    """ Delta-kick perturbation.

    Parameters
    ----------
    strength
        Strength of the perturbation in the frequency domain.
    """

    def __init__(self,
                 strength: float):
        self.strength = strength

    def amplitude(self,
                  times: NDArray[np.float64]) -> NDArray[np.float64]:
        dt = self.timestep(times)
        amplitudes = np.abs(times) < 1e-3 * dt   # 1 if zero, else 0

        return self.strength / dt * amplitudes

    def fourier(self,
                times: NDArray[np.float64],
                padnt: int | None = None) -> NDArray[np.complex128]:
        nw = len(self.frequencies(times, padnt))  # Length of frequencies grid
        return self.strength * np.ones(nw)  # type: ignore

    def normalize_frequency_response(self,
                                     data: NDArray[np.float64],
                                     times: NDArray[np.float64],
                                     padnt: int,
                                     axis: int = -1) -> NDArray[np.complex128]:
        data_w = np.fft.rfft(data, n=padnt) * self.timestep(times)
        # The strength is specified in the frequency domain, so the timestep is included in strength
        return data_w / self.strength

    def normalize_time_response(self,
                                data: NDArray[np.float64],
                                times: NDArray[np.float64],
                                axis: int = -1) -> NDArray[np.float64]:
        # The strength is specified in the frequency domain, hence no multiplication by timestep
        return data / self.strength  # type: ignore

    def todict(self) -> dict[str, Any]:
        return {'name': 'deltakick', 'strength': self.strength}

    def __str__(self) -> str:
        return f'Delta-kick perturbation (strength {self.strength:.1e})'


class PulsePerturbation(Perturbation):

    """ Perturbation as a time-dependent function.

    Parameters
    ----------
    pulse
        Object representing the pulse.
    """

    def __init__(self,
                 pulse: Laser | dict):
        self.pulse = create_laser(pulse)

    def amplitude(self,
                  times: NDArray[np.float64]) -> NDArray[np.float64]:
        return self.pulse.strength(times)

    def fourier(self,
                times: NDArray[np.float64],
                padnt: int | None = None) -> NDArray[np.complex128]:
        pulse_t = self.amplitude(times)
        if padnt is None:
            padnt = len(times)
        return np.fft.rfft(pulse_t, n=padnt) * self.timestep(times)

    def normalize_frequency_response(self,
                                     data: NDArray[np.float64],
                                     times: NDArray[np.float64],
                                     padnt: int,
                                     axis: int = -1) -> NDArray[np.complex128]:
        data = data.swapaxes(axis, -1)  # Put the time dimension last
        thresh = 0.005  # Threshold for filtering where perturbation is zero

        # Fourier transform of perturbation
        perturb_t = self.pulse.strength(times)
        perturb_w = np.fft.rfft(perturb_t, n=padnt)

        # Fourier transform of data
        data_w = np.fft.rfft(data, n=padnt)

        # Mask where perturbation is below threshold
        flt_w = np.abs(perturb_w) > thresh * np.abs(perturb_w).max()
        data_w[..., ~flt_w] = 0

        # Divide by the perturbation
        data_w[..., flt_w] /= perturb_w[flt_w]

        # Move back the time/frequency dimension
        data_w = data_w.swapaxes(axis, -1)

        return data_w

    def todict(self) -> dict[str, Any]:
        try:
            return self.pulse.todict()
        except AttributeError:
            return {'name': self.pulse.__class__.__name__}

    def __str__(self) -> str:
        lines: list[str] = []
        width = 50
        for key, value in self.todict().items():
            line = f'{key}: {value}'
            if len(lines) == 0:
                lines.append(line)
                continue
            if len(lines[-1]) + len(line) + 2 < width:
                lines[-1] = lines[-1] + ', ' + line
            else:
                lines.append(line)
        return '\n'.join(lines)



from .density import DensityCalculator
from .dipole import DipoleCalculator
from .energy import EnergyCalculator
from .hotcarriers import HotCarriersCalculator

__all__ = [
    'DensityCalculator',
    'DipoleCalculator',
    'EnergyCalculator',
    'HotCarriersCalculator',
]



from __future__ import annotations

from typing import Any, Generator
from pathlib import Path

import numpy as np
from numpy.typing import NDArray

from ase.io.ulm import Reader
from gpaw.mpi import world

from .typing import GPAWCalculator
from .utils import gauss_ij, Logger
from .voronoi import VoronoiWeights, atom_projections_to_numpy


class DOSCalculator:

    r""" Density of states (:term:`DOS`) and partial DOS (:term:`PDOS`) calculator.

    Calculates DOS

    .. math::

       \mathrm{DOS}(\varepsilon) = \Sigma_n G(\varepsilon - \varepsilon_n)

    and PDOS

    .. math::

       \mathrm{PDOS}(\varepsilon) = \Sigma_n w_{nn} G(\varepsilon - \varepsilon_n)

    where :math:`\varepsilon_n` are Kohn-Sham energies and :math:`G(\varepsilon - \varepsilon_n)`
    a Gaussian broadening function

    .. math::

       G(\varepsilon - \varepsilon_n)
        = \frac{1}{\sqrt{2 \pi \sigma^2}}
        \exp\left(-\frac{
            \left(\varepsilon_i - \varepsilon\right)^2
        }{
            2 \sigma^2
        }\right).

    The Voronoi weights are defined

    .. math::

        W_{nn}
        = \left<\psi_n|\hat{w}|\psi_{n}\right>
        = \int w(\boldsymbol{r}) \psi_n^*(\boldsymbol{r}) \psi_{n}(\boldsymbol{r}) d\boldsymbol{r}

    where the operator :math:`\hat{w} = w(\boldsymbol{r})` is 1 in the Voronoi
    region of the atomic projections and 0 outside.

    Parameters
    ----------
    eigenvalues
        List of eigenvalues :math:`\varepsilon_n` (relative to Fermi level).
    fermilevel
        Fermi level.
    voronoi
        Voronoi weights object defining the atomic projections for :term:`PDOS`.

        Leave out if only :term:`DOS` is desired.
    energies
        Array of energies (in units of eV) for which the broadened :term:`PDOS` is computed.
    sigma
        Gaussian broadening width :math:`\sigma` in units of eV.
    zerofermi
        Eigenvalues relative to Fermi level if `True`, else relative to vacuum
    """

    def __init__(self,
                 eigenvalues: list[float] | NDArray[np.float64],
                 fermilevel: float,
                 voronoi: VoronoiWeights | None,
                 energies: list[float] | NDArray[np.float64],
                 sigma: float,
                 zerofermi: bool = False):
        self._voronoi = voronoi
        self._eig_n = np.array(eigenvalues)
        self._fermilevel = fermilevel
        self._energies = np.array(energies)
        self._sigma = sigma
        self._zerofermi = zerofermi
        if voronoi is None:
            self._log = Logger()
        else:
            self._log = voronoi.log

    @property
    def voronoi(self) -> VoronoiWeights | None:
        """ Voronoi weights object. """
        return self._voronoi

    @property
    def log(self) -> Logger:
        """ Logger. """
        return self._log

    @property
    def zero(self) -> float:
        if self._zerofermi:
            return self._fermilevel
        else:
            return 0

    @property
    def eig_n(self) -> NDArray[np.float64]:
        return self._eig_n - self.zero

    @property
    def energies(self) -> NDArray[np.float64]:
        """ Energy grid in units of eV. """
        return self._energies

    @property
    def sigma(self) -> float:
        """ Gaussian broadening width in units of eV. """
        return self._sigma

    def calculate_dos(self) -> NDArray[np.float64] | None:
        """ Calculate :term:`DOS`.

        Returns
        ------
        DOS on the root rank, None on other ranks.
        """
        if world.rank != 0:
            return None

        # Construct gaussians
        gauss_en = gauss_ij(self.energies, self.eig_n, self.sigma)
        dos_e = np.sum(gauss_en, axis=1)
        self.log('Computed DOS', who='Calculator', comm=world, flush=True)
        return dos_e

    def icalculate_pdos(self) -> Generator[dict[str, NDArray[np.float64] | None], None, None]:
        """ Calculate :term:`PDOS` for each of the atomic projections in the :attr:`voronoi` object..

        Yields
        ------
            Once per set of Voronoi weights a dictionary with keys
                * `weight_n` - Array of dimensions `(Nn)` of projections. `None` on non-root ranks.
                * `pdos_e` - Broadened PDOS. `None` on non-root ranks.
        """
        if self.voronoi is None:
            raise ValueError('Voronoi must be given to the calculator')

        if world.rank == 0:
            # Construct gaussians
            gauss_en = gauss_ij(self.energies, self.eig_n, self.sigma)
            self.log('Computed gaussians', who='Calculator', comm=world, flush=True)

        for i, weight_nn in enumerate(self.voronoi):
            if world.rank == 0:
                assert weight_nn is not None
                weight_n = weight_nn.diagonal()
                pdos_e = gauss_en @ weight_n
                self.log(f'Computed PDOS for projection {self.voronoi.atom_projections[i]}',
                         who='Calculator', comm=world, flush=True)
                yield dict(weight_n=weight_n, pdos_e=pdos_e)
            else:
                yield dict(weight_n=None, pdos_e=None)

    @classmethod
    def from_gpw(cls,
                 gpw_file: Path | str,
                 voronoi: VoronoiWeights | None,
                 energies: list[float] | NDArray[np.float64],
                 sigma: float,
                 zerofermi: bool = False):
        r"""
        Initialize from `.gpw` file.

        Parameters
        ----------
        gpw_file
            File name of GPAW ground state file.
        voronoi
            Voronoi weights object defining the atomic projections for :term:`PDOS`.

            Leave out if only :term:`DOS` is desired.
        energies
            Array of energies (in units of eV) for which the broadened :term:`PDOS` is computed.
        sigma
            Gaussian broadening width :math:`\sigma` in units of eV.
        zerofermi
            Eigenvalues relative to Fermi level if `True`, else relative to vacuum.
        """

        reader = Reader(gpw_file)
        eig_skn = reader.wave_functions.eigenvalues
        # Assume only one spin channel and one k point
        assert eig_skn.shape[:2] == (1, 1), 'Many spins or kpoints'
        eig_n = eig_skn[0, 0]
        fermilevel = reader.wave_functions.fermi_levels[0]

        return cls(eig_n, fermilevel, voronoi, energies, sigma, zerofermi)

    @classmethod
    def from_calc(cls,
                  calc: GPAWCalculator,
                  voronoi: VoronoiWeights | None,
                  energies: list[float] | NDArray[np.float64],
                  sigma: float,
                  zerofermi: bool = False):
        r"""
        Initialize from GPAW calculator.

        Parameters
        ----------
        calc
            GPAW calculator.
        voronoi
            Voronoi weights object defining the atomic projections for :term:`PDOS`.

            Leave out if only :term:`DOS` is desired.
        energies
            Array of energies (in units of eV) for which the broadened :term:`PDOS` is computed.
        sigma
            Gaussian broadening width :math:`\sigma` in units of eV.
        zerofermi
            Eigenvalues relative to Fermi level if `True`, else relative to vacuum.
        """
        eig_n = calc.get_eigenvalues()
        fermilevel = calc.get_fermi_level()

        return cls(eig_n, fermilevel, voronoi, energies, sigma, zerofermi)

    def calculate_dos_and_write(self,
                                out_fname: Path | str,
                                write_extra: dict[str, Any] = dict()):
        """ Calculate the :term:`DOS` and write to file.

        The DOS is saved in a numpy archive if the file extension is `.npz`
        or in a comma separated file if the file extension is `.dat`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        write_extra
            Dictionary of additional data written to numpy archive (ignored for `.dat`) files.
        """
        from .writers.dos import write_density_of_states

        # Calculate
        if world.rank == 0:
            dos_e = self.calculate_dos()

        # Write to file on root
        if world.rank > 0:
            world.barrier()
            return
        assert dos_e is not None

        out_fname = str(out_fname)
        if out_fname[-4:] == '.npz':
            write: dict[str, Any] = dict(
                energy_e=self.energies,
                dos_e=dos_e,
                sigma=self.sigma,
                zerofermi=self._zerofermi,
            )
            write.update(sigma=self.sigma, zerofermi=self._zerofermi)
            write.update(write_extra)

            np.savez_compressed(out_fname, **write)
        elif out_fname[-4:] == '.dat':
            write_density_of_states(out_fname=out_fname,
                                    energies=self.energies,
                                    dos=dos_e,
                                    sigma=self.sigma,
                                    zerofermi=self._zerofermi)
        else:
            raise ValueError(f'output-file must have ending .npz or .dat, is {out_fname}')

        self.log(f'Written {out_fname}', flush=True, who='Calculator', rank=0)
        world.barrier()

    def calculate_pdos_and_write(self,
                                 out_fname: Path | str,
                                 write_extra: dict[str, Any] = dict(),
                                 write_extra_from_voronoi: bool = False):
        """ Calculate the :term:`PDOS` and write to file.

        The PDOS is saved in a numpy archive if the file extension is `.npz`
        or in a comma separated file if the file extension is `.dat`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        write_extra
            Dictionary of additional data written to numpy archive (ignored for `.dat`) files.
        write_extra_from_voronoi
            If true, and voronoi is a ULM reader, extra key-value pairs are read from
            voronoi and written to the `.npz` file (ignored for `.dat`) files.
        """
        from .writers.dos import write_partial_density_of_states

        if self.voronoi is None:
            raise ValueError('Voronoi must be given to the calculator')

        # Calculate
        if world.rank == 0:
            pdos_ei = np.zeros((len(self.energies), len(self.voronoi)))
        for i, ret in enumerate(self.icalculate_pdos()):
            if world.rank != 0:
                continue
            pdos_ei[:, i] = ret['pdos_e']

        # Write to file on root
        if world.rank > 0:
            world.barrier()
            return

        out_fname = str(out_fname)
        if out_fname[-4:] == '.npz':
            write: dict[str, Any] = dict(
                energy_e=self.energies,
                pdos_ei=pdos_ei,
                atom_projections=atom_projections_to_numpy(self.voronoi.atom_projections),
                sigma=self.sigma,
                zerofermi=self._zerofermi,
            )
            write.update(sigma=self.sigma, zerofermi=self._zerofermi)
            if write_extra_from_voronoi:
                write.update(self.voronoi.saved_fields)
            write.update(write_extra)

            np.savez_compressed(out_fname, **write)
        elif out_fname[-4:] == '.dat':
            write_partial_density_of_states(out_fname=out_fname,
                                            energies=self.energies,
                                            pdos=pdos_ei,
                                            atom_projections=self.voronoi.atom_projections,
                                            sigma=self.sigma,
                                            zerofermi=self._zerofermi)
        else:
            raise ValueError(f'output-file must have ending .npz or .dat, is {out_fname}')

        self.log(f'Written {out_fname}', flush=True, who='Calculator', rank=0)
        world.barrier()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Iterator, Sequence, Union

import numpy as np
from numpy.typing import NDArray

from gpaw import GPAW
from gpaw.analyse.wignerseitz import wignerseitz
from gpaw.io import Writer
from gpaw.mpi import broadcast, world
from gpaw.utilities.tools import tri2full

from .utils import Logger, parulmopen, ParallelMatrix
from .typing import Communicator, GPAWCalculator


AtomProjectionsType = Sequence[Union[list[int], NDArray[np.float64]]]  # | breaks on py3.9


class VoronoiWeights(ABC):

    """ Abstract base class for Voronoi weights in Kohn-Sham basis. """

    def __init__(self,
                 comm: Communicator | None = None,
                 log: Logger | None = None):
        if comm is None:
            comm = world
        if log is None:
            log = Logger()

        self._comm = comm
        self._log = log

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass

    def __len__(self) -> int:
        """ Number of projections. """
        return self.nproj

    @abstractmethod
    def __iter__(self) -> Iterator[NDArray[np.float64] | None]:
        """ Iteratively yield Voronoi weights for each projection.

        Yields
        ------
        Matrix of Voronoi weights on root rank, `None` on other ranks.
        """
        raise NotImplementedError

    def __str__(self) -> str:
        lines = [f'{self.__class__.__name__} for ground state with {self.nn} bands',
                 f'{self.nproj} projections:']

        for i, atoms in enumerate(self.atom_projections):
            atomsstr = str(atoms)
            if len(atomsstr) > 50:
                atomsstr = atomsstr[:47] + '...'
            lines.append(f'  #{i:<3.0f}: On atoms {atomsstr}')

        return '\n'.join(lines)

    @property
    def root(self) -> bool:
        """ Whether this rank is the root rank. """
        return self.comm.rank == 0

    @property
    def log(self) -> Logger:
        """ Logger object. """
        return self._log

    @property
    @abstractmethod
    def atom_projections(self) -> AtomProjectionsType:
        """ Atom projections. """
        raise NotImplementedError

    @property
    @abstractmethod
    def nn(self) -> int:
        """ Number of bands. """
        raise NotImplementedError

    @property
    def nproj(self) -> int:
        """ Number of projections. """
        return len(self.atom_projections)

    @property
    def comm(self) -> Communicator:
        """ MPI Communicator. """
        return self._comm

    @property
    @abstractmethod
    def saved_fields(self) -> dict[str, Any]:
        """ Saved data fields associated with the object.

        If this object is read from a ULM file there might be some extra
        data in that file. Return that data, or an empty dict if there is none.
        """
        raise NotImplementedError


class VoronoiWeightCalculator(VoronoiWeights):

    r"""Calculate Voronoi weights in the basis of ground state Kohn-Sham orbitals.

    For each atomic projection, calculates

    .. math::

        W_{nn'}
        = \left<\psi_n | \hat{w} | \psi_{n'}\right>
        = \left<\tilde{\psi}_n | \hat{w} | \tilde{\psi}_{n'}\right>
        + \sum_{aij} \left<\tilde{\psi}_n | \tilde{p}_i^a\right>
        \Delta S_{ij}^a \left<\tilde{p}_i^a | \tilde{\psi}_{n'}\right>

    where the operator :math:`\hat{w} = w(\vec{r})` is 1 in the
    Voronoi region of atoms and 0 outside.
    The sum over atoms is restricted to atoms in the region :math:`w(\vec{r}) = 1`.

    The PAW projectors

    .. math::

        P_{ni}^a = \left<\tilde{p}_i^a | \tilde{\psi}_n \right>

    and overlap matrix

    .. math::

        \Delta S_{ij}^a
        = \left<\phi_i^a|\phi_j^a\right>
        - \left<\tilde{\phi}_i^a|\tilde{\phi}_j^a\right>

    are obtained from the GPAW calculator and the smooth weight matrix is obtained by
    a basis change of the smooth LCAO weights

    .. math::

        \left<\tilde{\psi}_n | \hat{w} | \tilde{\psi}_{n'}\right> =
        \sum_{\mu\nu} C_{n\mu} C_{n'\nu}
        \left<\tilde{\phi}_\mu | \hat{w} | \tilde{\phi}_{\nu}\right>.

    Parameters
    ----------
    voronoi_lcao
        Object that calculates or loads the LCAO weights from file.
    """

    _voronoi_lcao: VoronoiLCAOWeights

    def __init__(self,
                 voronoi_lcao: VoronoiLCAOWeights):
        assert isinstance(voronoi_lcao, VoronoiLCAOWeights)
        self._voronoi_lcao = voronoi_lcao

        super().__init__(comm=voronoi_lcao.comm, log=voronoi_lcao.log)

    def __iter__(self) -> Iterator[NDArray[np.float64] | None]:
        for proj_atoms, weight_MM in zip(self.atom_projections, self.voronoi_lcao):
            nM = self.voronoi_lcao.nM
            nn = self.voronoi_lcao.nn

            # Transform to KS basis C_nM @ weight_MM @ C_nM.T
            w_MM = ParallelMatrix((nM, nM), np.float64, comm=self.comm,
                                  array=weight_MM)
            C_nM = ParallelMatrix((nn, nM), np.float64, comm=self.comm,
                                  array=self.voronoi_lcao.C_nM)

            _v_nn = C_nM @ w_MM @ C_nM.T
            self.log('Transformed weights to KS basis', flush=True, who='Voronoi', rank=0)

            # Calculate PAW corrections on the root rank
            if self.root:
                dS_aii = self.voronoi_lcao.dS_aii
                P_ani = self.voronoi_lcao.P_ani
                v_nn = _v_nn.array
                assert dS_aii is not None
                assert P_ani is not None

                for a, P_ni in P_ani.items():
                    if a not in proj_atoms:
                        continue
                    v_nn += P_ni @ dS_aii[a] @ P_ni.T

                self.log('Added PAW corrections to weights', flush=True, who='Voronoi', rank=0)

                yield v_nn
            else:
                yield None

    @property
    def voronoi_lcao(self) -> VoronoiLCAOWeights:
        return self._voronoi_lcao

    @property
    def atom_projections(self) -> AtomProjectionsType:
        return self.voronoi_lcao.atom_projections

    @property
    def nn(self) -> int:
        return self.voronoi_lcao.nn

    @property
    def nM(self) -> int:
        return self.voronoi_lcao.nM

    @property
    def comm(self) -> Communicator:
        return self.voronoi_lcao.comm

    @property
    def saved_fields(self) -> dict[str, Any]:
        return dict()

    def calculate_and_write(self,
                            out_fname: str,
                            write_extra: dict[str, Any] = dict()):
        """ Calculate the Voronoi weights in the Kohn-Sham basis and write to file.

        The weights are saved in a numpy archive if the file extension is `.npz`
        or in a ULM file if the file extension is `.ulm`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        write_extra
            Dictionary of additional data written to the file.
        """
        to_be_written = dict()
        if world.rank == 0:
            to_be_written.update(self.saved_fields)
            to_be_written.update(write_extra)

        if out_fname.endswith('.npz'):
            # Calculate weights
            weight_inn = list(self)

            # Write on root ranks
            if not self.root:
                return

            to_be_written['atom_projections'] = atom_projections_to_numpy(self.atom_projections)
            np.savez(out_fname, weight_inn=np.array(weight_inn), **to_be_written)
        elif out_fname.endswith('.ulm'):
            with Writer(out_fname, world, mode='w', tag='Voronoi') as writer:
                writer.write(version=1)
                writer.write('atom_projections', self.atom_projections)
                writer.write(**to_be_written)

                writer.add_array('weight_inn', (self.nproj, self.nn, self.nn), dtype=float)

                # Calculate weights
                for weight_nn in self:
                    # Write on root (DummyWriter on other ranks)
                    writer.fill(weight_nn)
        else:
            raise ValueError(f'output-file must have ending .npz or .ulm, is {out_fname}')
        self.log(f'Weights written to {out_fname}', flush=True, who='Voronoi', rank=0)
        world.barrier()

    @classmethod
    def from_gpw(cls,
                 atom_projections: AtomProjectionsType,
                 gpw_file: str,
                 voronoi_grid: VoronoiGrid | None | str = None,
                 comm: Communicator | None = None) -> VoronoiWeightCalculator:
        """
        Set up calculation from GPAW file.

        Parameters
        ----------
        atom_projections
            List of atom groups. Each atom group is a list of integers (of any length).
        gpw_file
            File name of GPAW ground state file.
        voronoi_grid
            Voronoi grid, or `None` to calculate it, or a file name to read it from file.
        comm
            Communicator.
        """
        voronoi_lcao = VoronoiLCAOWeightCalculator(
                atom_projections=atom_projections,
                gpw_file=gpw_file,
                comm=comm)
        return cls(voronoi_lcao)

    @classmethod
    def from_lcao_file(cls,
                       voronoi_lcao_file: str,
                       comm: Communicator | None = None) -> VoronoiWeightCalculator:
        """
        Set up calculation from file of already computed weights in LCAO basis.

        Parameters
        ----------
        voronoi_lcao_file
            File name of file with Voronoi weights in LCAO basis.
        comm
            Communicator.
        """
        voronoi_lcao = VoronoiLCAOReader(voronoi_lcao_file, comm=comm)
        return cls(voronoi_lcao)


class VoronoiReader(VoronoiWeights):

    """ Read Voronoi weights from ulm file.

    Parameters
    ----------
    ulm_fname
        File name.
    comm
        GPAW MPI communicator object. Defaults to world.
    """

    _nn: int
    _atom_projections: AtomProjectionsType

    def __init__(self,
                 ulm_fname: str,
                 comm: Communicator | None = None):
        super().__init__(comm=comm)

        self.ulm_fname = ulm_fname
        self.reader = parulmopen(self.ulm_fname, self.comm)

        # Read size
        if self.root:
            weight_inn = self.reader.proxy('weight_inn')
            assert weight_inn.dtype is np.dtype(float)
            assert weight_inn.shape[1] == weight_inn.shape[2]
            atom_projections = self.reader.atom_projections
            nn = weight_inn.shape[1]
            brdcast = (atom_projections, nn)
        else:
            brdcast = None

        brdcast = broadcast(brdcast, root=0, comm=self.comm)
        self._atom_projections, self._nn = brdcast  # type: ignore

    @property
    def atom_projections(self) -> AtomProjectionsType:
        return self._atom_projections

    @property
    def nn(self) -> int:
        return self._nn

    def __exit__(self, exc_type, exc_value, tb):
        self.reader.close()

    def __iter__(self) -> Iterator[NDArray[np.float64] | None]:
        for i in range(len(self)):
            if self.root:
                weight_nn = self.reader.proxy('weight_inn', i)[:]
            else:
                weight_nn = None

            yield weight_nn

    @property
    def saved_fields(self) -> dict[str, Any]:
        """ Saved data fields associated with the object

        If this object is read from a ULM file there might be some extra
        data in that file. Return that data, or an empty dict if there is none.
        """
        data = {key: getattr(self.reader, key) for key in self.reader.keys()
                if key not in ['weight_inn', 'atom_projections']}

        return data


class EmptyVoronoiWeights(VoronoiWeights):

    """ Object representing a lack of weights. """

    def __iter__(self):
        return
        yield

    @property
    def atom_projections(self):
        return []

    @property
    def nn(self):
        return 0

    @property
    def saved_fields(self):
        return {}


class VoronoiLCAOWeights(ABC):

    """ Abstract base class for Voronoi weights in LCAO basis. """

    _atom_projections: AtomProjectionsType
    _comm: Communicator
    _log: Logger
    _dS_aii: Any
    _P_ani: Any
    _C_nM: Any

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass

    def __len__(self) -> int:
        """ Return the number of projections. """
        return self.nproj

    @abstractmethod
    def __iter__(self) -> Iterator[NDArray[np.float64] | None]:
        """ Iteratively yield Voronoi weights in the LCAO basis for each projection.

        Yields
        ------
        Matrix of Voronoi weights on root rank, `None` on other ranks.
        """
        raise NotImplementedError

    @property
    def log(self) -> Logger:
        return self._log

    def log_parallel(self, *args, **kwargs) -> Logger:
        """ Log message with communicator information. """
        return self._log(*args, **kwargs, comm=self.comm, who='Voronoi')

    @property
    @abstractmethod
    def nn(self) -> int:
        """ Number of bands. """
        raise NotImplementedError

    @property
    @abstractmethod
    def nM(self) -> int:
        """ Number of basis functions. """
        raise NotImplementedError

    @property
    def nproj(self) -> int:
        """ Number of atomic projections """
        return len(self.atom_projections)

    @property
    def atom_projections(self) -> AtomProjectionsType:
        """ List of atom projections. """
        return self._atom_projections

    @property
    def calc(self) -> GPAWCalculator | None:
        """ GPAW calculator instance. """
        return None

    @property
    def comm(self) -> Communicator:
        """ MPI Communicator. """
        return self._comm

    @property
    def root(self) -> bool:
        """ Whether this rank is the root rank. """
        return self.comm.rank == 0

    @property
    def C_nM(self) -> NDArray[np.float64] | None:
        """ LCAO wave function coefficients on root rank, `None` on other ranks. """
        return self._C_nM if self.root else None

    @property
    def P_ani(self) -> dict[int, NDArray[np.float64]] | None:
        r""" PAW projectors :math:`P_{ni}^a` on the root rank, `None` on other ranks.

        .. math::

            P_{ni}^a = \left<\tilde{p}_i^a | \tilde{\psi}_n \right>
        """
        return self._P_ani if self.root else None

    @property
    def dS_aii(self) -> dict[int, NDArray[np.float64]] | None:
        r""" Overlap matrix PAW corrections :math:`\Delta S_{ij}^a` on the root rank, `None` on other ranks.

        .. math::

            \Delta S_{ij}^a
            = \left<\phi_i^a|\phi_j^a\right>
            - \left<\tilde{\phi}_i^a|\tilde{\phi}_j^a\right>
        """
        return self._dS_aii if self.root else None

    @property
    @abstractmethod
    def saved_fields(self) -> dict[str, Any]:
        """ Saved data fields associated with the object.

        If this object is read from a ULM file there might be some extra
        data in that file. Return that data, or an empty dict if there is none.
        """
        raise NotImplementedError

    @property
    def arrays(self) -> dict[str, Any]:
        if self.comm.rank != 0:
            return {}

        arrays = dict(P_ani=self.P_ani,
                      C_nM=self.C_nM,
                      dS_aii=self.dS_aii)

        return arrays


class VoronoiLCAOReader(VoronoiLCAOWeights):

    """ Read Voronoi weights in the LCAO basis from ULM file.

    Parameters
    ----------
    ulm_fname
        File name.
    comm
        GPAW MPI communicator object. Defaults to world.
    """

    def __init__(self,
                 ulm_fname: str,
                 comm=None):
        self.ulm_fname = ulm_fname

        if comm is None:
            comm = world

        self._log = Logger()
        self._comm = comm
        self.reader = parulmopen(self.ulm_fname, self.comm)

        self._dS_aii: dict[int, NDArray[np.float64]] | None
        self._P_ani: dict[int, NDArray[np.float64]] | None
        self._C_nM = NDArray[np.float64] | None

        # Read arrays on root rank
        if self.root:
            weight_iMM = self.reader.proxy('weight_iMM')
            assert weight_iMM.dtype is np.dtype(float)
            assert weight_iMM.shape[1] == weight_iMM.shape[2]
            self._dS_aii = self.reader.dS_aii
            self._P_ani = self.reader.P_ani
            self._C_nM = self.reader.C_nM[:]
            atom_projections = self.reader.atom_projections
            brdcast = (atom_projections, ) + self._C_nM.shape
        else:
            brdcast = None
            self._C_nM = None
            self._P_ani = None
            self._dS_aii = None

        # Broadcast atom projections and number of bands and basis functions
        brdcast = broadcast(brdcast, root=0, comm=self.comm)
        self._atom_projections, self._nn, self._nM = brdcast  # type: ignore

    def __exit__(self, exc_type, exc_value, tb):
        self.reader.close()

    @property
    def nn(self) -> int:
        return self._nn

    @property
    def nM(self) -> int:
        return self._nM

    def __iter__(self) -> Iterator[NDArray[np.float64]]:
        for i in range(len(self)):
            if self.root:
                weight_MM = self.reader.proxy('weight_iMM', i)[:]
            else:
                weight_MM = None

            yield weight_MM

    @property
    def saved_fields(self) -> dict[str, Any]:
        data = {key: getattr(self.reader, key) for key in self.reader.keys()
                if key not in ['weight_iMM', 'atom_projections',
                               'dS_aii', 'P_ani', 'C_nM']}

        return data


class VoronoiLCAOWeightCalculator(VoronoiLCAOWeights):

    r"""Calculate Voronoi weights in LCAO basis.

    For each atomic projection, calculates

    .. math::

        \tilde{v}_{\mu\nu}
        = \left<\tilde{\phi}_{\mu} | \hat{w} | \tilde{\phi}_{\nu}\right>
        = \int w(\vec{r}) \tilde{\phi}_{\mu}^*(\vec{r}) \tilde{\phi}_{\nu}(\vec{r}) d\vec{r}

    where the operator :math:`\hat{w} = w(\vec{r})` is 1 in the
    Voronoi region of the atomic projections and 0 outside, and :math:`\tilde{\phi}_{\nu}` are
    the smooth LCAO basis functions.

    Parameters
    ----------
    atom_projections
        List of atom groups. Each atom group is a list of integers (of any length).
    gpw_file
        File name of GPAW ground state file.
    voronoi_grid
        Voronoi grid, or `None` to calculate it, or a file name to read it from file.
    domain
        Domain size.
    comm
        Communicator.
    """

    _calc: GPAWCalculator

    def __init__(self,
                 atom_projections: AtomProjectionsType,
                 gpw_file: str,
                 voronoi_grid: VoronoiGrid | None | str = None,
                 domain: int = -1,
                 comm: Communicator | None = None):
        assert all([isinstance(proj_atoms, list) or isinstance(proj_atoms, np.ndarray)
                    for proj_atoms in atom_projections])
        self._atom_projections = atom_projections

        if comm is None:
            comm = world

        self._comm = comm
        self._log = Logger()

        if voronoi_grid is None:
            voronoi_grid = VoronoiGridCalculator()
        if isinstance(voronoi_grid, str):
            voronoi_grid = VoronoiGridReader(voronoi_grid)
        self.voronoi_grid = voronoi_grid

        if domain == -1:
            domain = self.comm.size

        # Set up GPAW calculator
        calc = GPAW(gpw_file, txt=None, communicator=self.comm, parallel={'domain': domain})
        calc.initialize_positions()
        self.log('Loaded and initialized GPAW', rank=0, flush=True, who='Voronoi')
        self._calc = calc  # type: ignore

        # Collect wave functions, projectors and overlap to the root rank
        C_nM = calc.wfs.collect_array('C_nM', k=0, s=0)
        P_ani = proj_as_dict_on_master(self.calc.wfs.kpt_u[0].projections, 0, self.nn)
        dS_aii = {a: setup.dO_ii for a, setup in enumerate(calc.wfs.setups)}

        self._C_nM = C_nM if self.root else None
        self._P_ani = P_ani if self.root else None
        self._dS_aii = dS_aii if self.root else None

    @property
    def nn(self) -> int:
        return self.calc.wfs.bd.nbands

    @property
    def nM(self) -> int:
        return self.calc.wfs.setups.nao

    @property
    def calc(self) -> GPAWCalculator:
        return self._calc

    def __iter__(self) -> Iterator[NDArray[np.float64] | None]:
        # Get the Voronoi grid
        a_G = self.voronoi_grid.a_G(self.calc, self.log)

        for proj_atoms in self.atom_projections:
            w_G = np.where(np.isin(a_G, proj_atoms), 1.0, 0.0)
            if self.calc.comms['b'].rank != 0:
                # The band comm ranks compute the same information
                return None

            # Calculate the weights on the domain communicator
            weight_MM = self.calc.wfs.basis_functions.calculate_potential_matrices(w_G)[0]
            tri2full(weight_MM)

            # Sum to the root rank
            self.calc.comms['d'].sum(weight_MM, root=0)

            self.log(f'Computed LCAO weights for projection {proj_atoms}', rank=0, flush=True, who='Voronoi')

            if self.root:
                yield weight_MM
            else:
                yield None

    @property
    def saved_fields(self) -> dict[str, Any]:
        return dict()

    def calculate_and_write(self,
                            out_fname: str,
                            write_extra: dict[str, Any] = dict()):
        """ Calculate the Voronoi weights in the LCAO basis and write to file.

        The weights are saved in a numpy archive if the file extension is `.npz`
        or in a ULM file if the file extension is `.ulm`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        write_extra
            Dictionary of additional data written to the file.
        """
        to_be_written = dict()
        if world.rank == 0:
            to_be_written.update(self.saved_fields)
            to_be_written.update(self.arrays)
            to_be_written.update(write_extra)

        if out_fname.endswith('.npz'):
            # Calculate weights
            weight_iMM = list(self)

            # Write on root ranks
            if self.root:
                return

            to_be_written['atom_projections'] = atom_projections_to_numpy(self.atom_projections)
            np.savez(out_fname, weight_iMM=np.array(weight_iMM), **to_be_written)
        elif out_fname.endswith('.ulm'):
            with Writer(out_fname, self.comm, mode='w', tag='Voronoi') as writer:
                writer.write(version=1)
                writer.write('atom_projections', self.atom_projections)
                writer.write(**to_be_written)

                writer.add_array('weight_iMM', (self.nproj, self.nM, self.nM), dtype=float)

                # Calculate weights
                for weight_MM in self:
                    # Write on root (DummyWriter on other ranks)
                    writer.fill(weight_MM)
        else:
            raise ValueError(f'output-file must have ending .npz or .ulm, is {out_fname}')
        self.log(f'Written weights in LCAO basis to {out_fname}', flush=True, who='Voronoi', rank=0)
        world.barrier()


class VoronoiGrid(ABC):

    @abstractmethod
    def a_G(self,
            calc: GPAWCalculator,
            log: Logger) -> NDArray[np.int_]:
        """ Voronoi grid (on the coarse grid of the GPAW calculator)
        distributed on the domain communicator.

        Each element in the grid is an integer corresponding to the closest atom.

        Parameters
        ----------
        GPAW calculator.
        """
        raise NotImplementedError

    def write(self,
              filename: str,
              calc: GPAWCalculator,
              log: Logger):
        """ Write grid to file.

        Parameters
        ----------
        file name
            File name.
        calc
            GPAW calculator.
        """
        filename = str(filename)

        # Calculate grid and collect to root
        big_a_G = calc.density.gd.collect(self.a_G(calc, log))
        if filename.endswith('.npz'):
            np.savez_compressed(filename, a_G=big_a_G)
        else:
            np.save(filename, big_a_G)
        log(f'Written Voronoi grid to {filename}', who='Voronoi', flush=True, rank=0)


class VoronoiGridReader(VoronoiGrid):

    """ Read Voronoi grid from file. """

    def __init__(self,
                 filename: str):
        self.filename = filename

    def a_G(self,
            calc: GPAWCalculator,
            log: Logger) -> NDArray[np.int_]:
        # Read the grid on the domain communicator root
        domain_comm = calc.comms['d']
        if domain_comm.rank == 0:
            if self.filename.endswith('.npz'):
                files = np.load(self.filename)
                big_a_G = files['a_G']
            else:
                big_a_G = np.load(self.filename)

            assert big_a_G.dtype == np.int16

        # Distribute grid across domain communicator
        gd = calc.density.gd
        a_G = gd.zeros(dtype=np.int16)
        gd.distribute(big_a_G if domain_comm.rank == 0 else None, a_G)

        log('Loaded Voronoi grid', rank=0, flush=True, who='Voronoi')

        return a_G


class VoronoiGridCalculator(VoronoiGrid):

    """ Calculate the Voronoi grid. """

    def a_G(self,
            calc: GPAWCalculator,
            log: Logger) -> NDArray[np.int_]:
        log('Computing Voronoi grid', rank=0, flush=True, who='Voronoi')
        atoms = calc.get_atoms()
        a_G = wignerseitz(calc.density.gd, atoms)
        a_G = a_G.astype(np.int16)
        log('Computed Voronoi grid', rank=0, flush=True, who='Voronoi')

        return a_G


def atom_projections_to_numpy(atom_projections: AtomProjectionsType) -> NDArray[np.int_]:
    Ni = len(atom_projections)
    if Ni == 0:
        Nj = 0
    else:
        Nj = max([len(proj_atoms) for proj_atoms in atom_projections])
    atom_projections_ij = np.full((Ni, Nj), -1, dtype=int)
    for i, proj_atoms in enumerate(atom_projections):
        na = len(proj_atoms)
        atom_projections_ij[i, :na] = proj_atoms

    return atom_projections_ij


def proj_as_dict_on_master(proj, n1: int, n2: int) -> dict[int, NDArray[np.float64]]:
    """ Collect the projectors to a dictionary on the root rank."""
    # In newer versions of GPAW this is proj.as_dict_on_master
    P_nI = proj.collect()
    if P_nI is None:
        return {}
    I1 = 0
    P_ani = {}
    for a, ni in enumerate(proj.nproj_a):
        I2 = I1 + ni
        P_ani[a] = P_nI[n1:n2, I1:I2]
        I1 = I2
    return P_ani



from __future__ import annotations

from pathlib import Path
import numpy as np
from numpy.typing import NDArray
from typing import Any, Generator, Sequence, Collection

from ase.units import Bohr
from gpaw import GPAW
from gpaw.grid_descriptor import GridDescriptor
from gpaw.lcaotddft.densitymatrix import get_density

from .base import BaseObservableCalculator
from ..typing import GPAWCalculator
from ..density_matrices.base import WorkMetadata
from ..density_matrices.frequency import FrequencyDensityMatrixMetadata
from ..density_matrices.time import ConvolutionDensityMatrixMetadata
from ..perturbation import PerturbationLike
from ..response import BaseResponse
from ..typing import ArrayIsOnRootRank, Array1D, DistributedArray
from ..utils import ResultKeys, Result, get_gaussian_pulse_values, ParallelMatrix


class DensityCalculator(BaseObservableCalculator):

    r""" Calculate densities in the time or frequency domain.

    The induced density (i.e. the density minus the ground state density) is to first
    order given by

    .. math::

        \delta n(\boldsymbol{r}) = -2 \sum_{ia}^\text{eh}
        n_{ia}(\boldsymbol{r}) \mathrm{Re}\:\delta\rho_{ia}

    plus PAW corrections, where :math:`n_{ia}(\boldsymbol{r})` is the density of
    ground state Kohn-Sham pair :math:`ia`

    .. math::

        n_{ia}(\boldsymbol{r}) = \psi^{(0)}_i(\boldsymbol{r}) \psi^{(0)}_a(\boldsymbol{r}).

    In the time domain, electrons and holes densities can be computed.

    .. math::

        \begin{align}
            n^\text{holes}(\boldsymbol{r}) &= \sum_{ii'}
            n_{ii'}(\boldsymbol{r}) \delta\rho_{ii'} \\
            n^\text{electrons}(\boldsymbol{r}) &= \sum_{aa'}
            n_{aa'}(\boldsymbol{r}) \delta\rho_{aa'}.
        \end{align}

    Refer to the documentation of
    :class:`HotCarriersCalculator <rhodent.calculators.HotCarriersCalculator>` for definitions
    of :math:`\delta\rho_{ii'}` and :math:`\delta\rho_{aa'}`.

    Parameters
    ----------
    gpw_file
        File name of GPAW ground state file.
    response
        Response object.
    filter_occ
        Filters for occupied states (holes). Provide a list of tuples (low, high)
        to compute the density of holes with energies within the interval low-high.
    filter_unocc
        Filters for unoccupied states (electrons). Provide a list of tuples (low, high)
        to compute the density of excited electrons with energies within the interval low-high.
    times
        Compute densities in the time domain, for these times (or as close to them as possible).
        In units of as.

        May not be used together with :attr:`frequencies` or :attr:`frequency_broadening`.
    pulses
        Compute densities in the time domain, in response to these pulses.
        If none, then no pulse convolution is performed.

        May not be used together with :attr:`frequencies` or :attr:`frequency_broadening`.
    frequencies
        Compute densities in the frequency domain, for these frequencies. In units of eV.

        May not be used together with :attr:`times` or :attr:`pulses`.
    frequency_broadening
        Compute densities in the frequency domain, with Gaussian broadening of this width.
        In units of eV.

        May not be used together with :attr:`times` or :attr:`pulses`.
    """

    def __init__(self,
                 gpw_file: str,
                 response: BaseResponse,
                 filter_occ: Sequence[tuple[float, float]] = [],
                 filter_unocc: Sequence[tuple[float, float]] = [],
                 *,
                 times: list[float] | Array1D[np.float64] | None = None,
                 pulses: Collection[PerturbationLike] | None = None,
                 frequencies: list[float] | Array1D[np.float64] | None = None,
                 frequency_broadening: float = 0,
                 ):
        super().__init__(response=response,
                         times=times,
                         pulses=pulses,
                         frequencies=frequencies,
                         frequency_broadening=frequency_broadening)
        self._occ_filters = [self._build_single_filter('o', low, high) for low, high in filter_occ]
        self._unocc_filters = [self._build_single_filter('u', low, high) for low, high in filter_unocc]

        self.log.start('load_gpaw')
        self._calc = GPAW(gpw_file, txt=None, communicator=self.calc_comm,
                          parallel={'domain': self.calc_comm.size})
        msg = f'Loaded/initialized GPAW in {self.log.elapsed("load_gpaw"):.1f}'
        self.log.start('init_gpaw')

        self.calc.initialize_positions()  # Initialize in order to calculate density
        msg += f'/{self.log.elapsed("init_gpaw"):.1f} s'
        if self.calc_comm.rank == 0:
            self.log_parallel(msg)
        self.ksd.density = self.calc.density

    @property
    def gdshape(self) -> tuple[int, int, int]:
        """ Shape of the real space grid.
        """
        shape = tuple(int(N) - 1 for N in self.N_c)
        return shape  # type: ignore

    @property
    def gd(self) -> GridDescriptor:
        """ Real space grid. """
        return self.ksd.density.finegd

    @property
    def N_c(self) -> NDArray[np.int_]:
        """ Number of points in each Cartesian direction of the grid.
        """
        return self.gd.N_c

    @property
    def cell_cv(self) -> NDArray[np.float64]:
        """ Cell vectors. """
        return self.gd.cell_cv * Bohr

    @property
    def occ_filters(self) -> list[slice]:
        """ List of energy filters for occupied states. """
        return self._occ_filters

    @property
    def unocc_filters(self) -> list[slice]:
        """ List of energy filters for unoccupied states. """
        return self._unocc_filters

    @property
    def calc(self) -> GPAWCalculator:
        """ GPAW calculator instance. """
        return self._calc  # type: ignore

    def get_result_keys(self,
                        yield_total: bool = True,
                        yield_electrons: bool = False,
                        yield_holes: bool = False) -> ResultKeys:
        noccf = len(self.occ_filters)
        nunoccf = len(self.unocc_filters)
        if (yield_electrons or yield_holes) and not self._is_time_density_matrices:
            raise ValueError('Electron or hole densities can only be computed in the time domain.')

        resultkeys = ResultKeys()
        if yield_total:
            resultkeys.add_key('rho_g', self.gdshape)

        if yield_holes:
            resultkeys.add_key('occ_rho_g', self.gdshape)
            if noccf > 0:
                resultkeys.add_key('occ_rho_rows_fg', (noccf, ) + self.gdshape)
                resultkeys.add_key('occ_rho_diag_fg', (noccf, ) + self.gdshape)

        if yield_electrons:
            resultkeys.add_key('unocc_rho_g', self.gdshape)
            if nunoccf > 0:
                resultkeys.add_key('unocc_rho_rows_fg', (nunoccf, ) + self.gdshape)
                resultkeys.add_key('unocc_rho_diag_fg', (nunoccf, ) + self.gdshape)

        return resultkeys

    @property
    def _need_derivatives_real_imag(self) -> tuple[list[int], bool, bool]:
        # Time domain: We only need the real part of the density matrix.
        # Frequency domain: We need the (complex) Fourier transform of
        #                   the real part of the density matrix.
        return ([0], True, False)

    def _find_limit(self,
                    lim: float) -> int:
        """ Find the first eigenvalue larger than :attr:`lim`.

        Parameters
        ----------
        lim
            Threshold value in units of eV.

        Returns
        -------
        The index of the first eigenvalue larger than :attr:`lim`.
        Returns `len(eig_n)` if :attr:`lim` is larger than all eigenvalues.
        """
        if lim > self.eig_n[-1]:
            return len(self.eig_n)
        return int(np.argmax(self.eig_n > lim))

    def _build_single_filter(self,
                             key: str,
                             low: float,
                             high: float) -> slice:
        imin, imax, amin, amax = self.ksd.ialims()

        if key == 'o':
            nlow = min(self._find_limit(low), imax) - imin
            nhigh = min(self._find_limit(high), imax) - imin
        elif key == 'u':
            nlow = min(self._find_limit(low), amax) - amin
            nhigh = min(self._find_limit(high), amax) - amin
        else:
            raise RuntimeError(f'Unknown key {key}. Key must be "o" or "u"')
        return slice(nlow, nhigh)

    def get_density(self,
                    rho_nn: DistributedArray,
                    nn_indices: str,
                    fltn1: slice | NDArray[np.bool_] = slice(None),
                    fltn2: slice | NDArray[np.bool_] = slice(None),
                    u: int = 0) -> DistributedArray:
        r""" Calculate a real space density from a density matrix in the Kohn-Sham basis.

        Parameters
        ----------
        rho_nn
            Density matrix :math:`\delta\rho_{ia}`, :math:`\delta\rho_{ii'}`, or
            :math:`\delta\rho_{aa'}`.
        nn_indices
            Indices describing the density matrices :attr:`rho_nn`. One of

            - `ia` for induced density :math:`\delta\rho_{ia'}`.
            - `ii` for holes density :math:`\delta\rho_{ii'}`.
            - `aa` for electrons density :math:`\delta\rho_{aa'}`.
        flt_n1
            Filter selecting rows of the density matrix.
        flt_n2
            Filter selecting columns of the density matrix.
        u
            k-point index.
        Returns
        -------
        Distributed array with the density in real space on the root rank.
        """
        imin, imax, amin, amax = self.ksd.ialims()
        if nn_indices not in ['ia', 'ii', 'aa']:
            raise ValueError(f'Parameter nn_indices must be either "ia", "ii" or "aa". Is {nn_indices}.')
        n1 = slice(imin, imax + 1) if nn_indices[0] == 'i' else slice(amin, amax + 1)
        n2 = slice(imin, imax + 1) if nn_indices[1] == 'i' else slice(amin, amax + 1)

        nn1, nn2 = n1.stop - n1.start, n2.stop - n2.start
        nM = self.ksd.C0_unM[0].shape[-1]

        if self.calc_comm.rank == 0:
            C0_nM = self.ksd.C0_unM[u]
            rho_n1n2 = ParallelMatrix((nn1, nn2), np.float64, comm=self.calc_comm,
                                      array=rho_nn[fltn1][:, fltn2])
            C0_n1M = ParallelMatrix((nn1, nM), np.float64, comm=self.calc_comm,
                                    array=C0_nM[n1][fltn1])
            C0_n2M = ParallelMatrix((nn2, nM), np.float64, comm=self.calc_comm,
                                    array=C0_nM[n2][fltn2])
        else:
            rho_n1n2 = ParallelMatrix((nn1, nn2), np.float64, comm=self.calc_comm)
            C0_n1M = ParallelMatrix((nn1, nM), np.float64, comm=self.calc_comm)
            C0_n2M = ParallelMatrix((nn2, nM), np.float64, comm=self.calc_comm)

        # Transform to LCAO basis C0_n1M.T @ rho_n1n2 @ C0_n2M
        self.log.start('transform_dm')

        rho_MM = (C0_n1M.T @ rho_n1n2 @ C0_n2M).broadcast()
        # assert np.issubdtype(rho_nn.dtype, float)
        rho_MM = 0.5 * (rho_MM + rho_MM.T)

        msg = f'Transformed DM and constructed density in {self.log.elapsed("transform_dm"):.1f}s'
        self.log.start('get_density')
        rho_g = get_density(rho_MM, self.calc.wfs, self.calc.density, u=u)
        msg += f'+{self.log.elapsed("get_density"):.1f}s'
        if self.calc_comm.rank == 0:
            self.log_parallel(msg, flush=True)

        big_rho_g = self.gd.collect(rho_g)

        if self.calc_comm.rank == 0:
            return big_rho_g
        else:
            return ArrayIsOnRootRank()

    def icalculate(self,
                   yield_total: bool = True,
                   yield_electrons: bool = False,
                   yield_holes: bool = False) -> Generator[tuple[WorkMetadata, Result], None, None]:
        """ Iteratively calculate results.

        Parameters
        ----------
        yield_total
            The results should include the total induced density.
        yield_holes
            The results should include the holes densities, optionally decomposed by `filter_occ`.
        yield_electrons
            The results should include the electrons densities, optionally decomposed by `filter_unocc`.

        Yields
        ------
        Tuple (work, result) on the root rank of the calculation communicator. \
        Does not yield on non-root ranks of the calculation communicator.

            work
                An object representing the metadata (time, frequency or pulse) for the work done.
            result
                Object containg the calculation results for this time, frequency or pulse.
        """
        noccf = len(self.occ_filters)
        nunoccf = len(self.unocc_filters)

        if (yield_electrons or yield_holes) and not self._is_time_density_matrices:
            raise ValueError('Electron or hole densities can only be computed in the time domain.')

        # Iterate over the pulses and times, or frequencies
        for work, dm in self.density_matrices:
            if self._is_time_density_matrices:
                # Real part contributes to density
                rho_ia = dm.rho_ia.real
            else:
                # Imaginary part gives absorption contribution
                rho_ia = -dm.rho_ia.imag

            self.log.start('calculate')

            # Non-root ranks on calc_comm will write empty arrays to result, but will not be yielded
            result = Result()

            if yield_total:
                result['rho_g'] = self.get_density(rho_ia.real, 'ia') * Bohr ** -3

            if yield_holes:
                # Holes
                M_ii = 0.5 * (dm.Q_ia @ dm.Q_ia.T + dm.P_ia @ dm.P_ia.T)

                result['occ_rho_g'] = self.get_density(M_ii, 'ii') * Bohr ** -3

                if noccf > 0:
                    result['occ_rho_rows_fg'] = np.array([self.get_density(M_ii, 'ii', fltn1=flt)
                                                          for flt in self.occ_filters]) * Bohr ** -3
                    result['occ_rho_diag_fg'] = np.array([self.get_density(M_ii, 'ii', fltn1=flt, fltn2=flt)
                                                          for flt in self.occ_filters]) * Bohr ** -3

            if yield_electrons:
                # Electrons
                M_aa = 0.5 * (dm.Q_ia.T @ dm.Q_ia + dm.P_ia.T @ dm.P_ia)

                result['unocc_rho_g'] = self.get_density(M_aa, 'aa') * Bohr ** -3

                if nunoccf > 0:
                    result['unocc_rho_rows_fg'] = np.array([self.get_density(M_aa, 'aa', fltn1=flt)
                                                            for flt in self.unocc_filters]) * Bohr ** -3
                    result['unocc_rho_diag_fg'] = np.array([self.get_density(M_aa, 'aa', fltn1=flt, fltn2=flt)
                                                            for flt in self.unocc_filters]) * Bohr ** -3
            if dm.rank > 0:
                continue

            self.log_parallel(f'Calculated density in {self.log.elapsed("calculate"):.2f}s '
                              f'for {work.desc}', flush=True)

            yield work, result

        if self.calc_comm.rank == 0:
            self.log_parallel('Finished calculating density contributions', flush=True)

    def calculate_and_write(self,
                            out_fname: str,
                            which: str | list[str] = 'induced',
                            write_extra: dict[str, Any] = dict()):
        """ Calculate density contributions.

        Densities are saved in a numpy archive, ULM file or cube file depending on
        whether the file extension is `.npz`, `.ulm`, or `.cube`.

        If the file extension is `.cube` then :attr:`out_fname` is taken to be a formatting string.

        The formatting string should be a plain string containing variable
        placeholders within curly brackets `{}`. It should not be confused with
        a formatted string literal (f-string).

        It acccepts the variables:

        * `{time}` - Time in units of as (time domain only).
        * `{freq}` - Frequency in units of eV (frequency domain only).
        * `{which}` - The :attr:`which` argument.
        * `{pulsefreq}` - Pulse frequency in units of eV (time domain only).
        * `{pulsefwhm}` - Pulse FWHM in units of fs (time domain only).

        Examples:

        * out_fname = `{which}_density_t{time:09.1f}.cube`.
        * out_fname = `{which}_density_w{freq:05.2f}.cube`.

        Parameters
        ----------
        out_fname
            File name of the resulting data file.
        which
            String, or list of strings specifying the types of density to compute:

            * `induced` - Induced density.
            * `holes` - Holes density (only allowed in the time domain).
            * `electrons` - Electrons density (only allowed in the time domain).
        write_extra
            Dictionary of extra key-value pairs to write to the data file.
        """
        from ..writers.density import DensityWriter, write_density
        from ..writers.writer import FrequencyResultsCollector, TimeResultsCollector

        cls = TimeResultsCollector if self._is_time_density_matrices else FrequencyResultsCollector

        if isinstance(which, str):
            which = [which]

        for which_key in which:
            if which_key in ['holes', 'electrons'] and not self._is_time_density_matrices:
                raise ValueError(f'Option which={which_key} not allowed in the frequency domain.')
            if which_key not in ['induced', 'holes', 'electrons']:
                raise ValueError(f'Option which={which} not recognized. '
                                 'Must be one of: induced, holes, electrons')

        calc_kwargs = dict(yield_total='induced' in which,
                           yield_holes='holes' in which,
                           yield_electrons='electrons' in which)

        keys = {'induced': 'rho_g',
                'holes': 'occ_rho_g',
                'electrons': 'unocc_rho_g'}

        out_fname = str(out_fname)
        if out_fname.endswith('.npz'):
            exclude = ['occ_rho_rows_fg', 'occ_rho_diag_fg', 'unocc_rho_rows_fg', 'unocc_rho_diag_fg']
            writer = DensityWriter(cls(self, calc_kwargs=calc_kwargs, exclude=exclude))
            writer.calculate_and_save_npz(out_fname=out_fname, write_extra=write_extra)
        elif out_fname.endswith('.ulm'):
            exclude = ['rho_g', 'occ_rho_rows_fg', 'occ_rho_diag_fg', 'unocc_rho_rows_fg', 'unocc_rho_diag_fg']
            writer = DensityWriter(cls(self, calc_kwargs=calc_kwargs, exclude=exclude))
            writer.calculate_and_save_ulm(out_fname=out_fname, write_extra=write_extra)
        elif out_fname.endswith('.cube'):
            atoms = self.calc.atoms
            for work, res in self.icalculate(**calc_kwargs):
                if self.calc_comm.rank > 0:
                    continue
                for which_key in which:
                    key = keys[which_key]
                    fname_kw: dict[str, float | str] = dict(which=which_key)
                    data = res[key]
                    if self._is_time_density_matrices:
                        assert isinstance(work, ConvolutionDensityMatrixMetadata)
                        fname_kw.update(time=work.time, **get_gaussian_pulse_values(work.pulse))
                    else:
                        assert isinstance(work, FrequencyDensityMatrixMetadata)
                        fname_kw.update(freq=work.freq)
                    fpath = Path(out_fname.format(**fname_kw))
                    fpath.parent.mkdir(parents=True, exist_ok=True)
                    write_density(str(fpath), atoms, data)
                    self.log_parallel(f'Written {fpath}', flush=True)

        else:
            raise ValueError(f'output-file must have ending .npz or .ulm, is {out_fname}')
